# (C) Copyright 2018- ECMWF.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

"""
A collection of tests for :any:`SymbolAttrs`, :any:`SymbolTable` and :any:`Scope`.
"""

import pytest

from loki.types import INTEGER, REAL, Scope, SymbolAttributes


def test_symbol_attributes():
    """
    Tests the attachment, lookup and deletion of arbitrary attributes from
    :any:`SymbolAttributes`
    """
    _type = SymbolAttributes('integer', a='a', b=True, c=None)
    assert _type.dtype == INTEGER
    assert _type.a == 'a'
    assert _type.b
    assert _type.c is None
    assert _type.foofoo is None

    _type.foofoo = 'bar'
    assert _type.foofoo == 'bar'

    delattr(_type, 'foofoo')
    assert _type.foofoo is None

    _type.b = None
    assert _type.b is None


def test_symbol_attributes_compare():
    """
    Test dedicated `type.compare` methods that allows certain
    attributes to be excluded from comparison.
    """
    someint = SymbolAttributes('integer', a='a', b=True, c=None)
    another = SymbolAttributes('integer', a='a', b=False, c=None)
    somereal = SymbolAttributes('real', a='a', b=True, c=None)

    assert not someint.compare(another)
    assert not another.compare(someint)
    assert someint.compare(another, ignore='b')
    assert another.compare(someint, ignore=['b'])
    assert not someint.compare(somereal)


def test_scope_setter():
    """ Test basic declaration and update behaviour of :any:`Scope` """
    scope = Scope()

    # Check basic type declaration
    scope.declare('a', dtype='integer', kind=4, intent='in')
    assert 'a' in scope.symbol_attrs
    assert scope.symbol_attrs['a'].dtype == INTEGER
    assert scope.symbol_attrs['a'].kind == 4
    assert scope.symbol_attrs['a'].intent == 'in'

    # Test erroneous and intentional re-declaration
    with pytest.raises(ValueError):
        scope.declare('a', dtype='real', kind=8)

    scope.declare('a', dtype='real', kind=8, fail=False)
    assert 'a' in scope.symbol_attrs
    assert scope.symbol_attrs['a'].dtype == REAL
    assert scope.symbol_attrs['a'].kind == 8
    assert not scope.symbol_attrs['a'].intent  # Wiped previous value

    # Check type declaration updates
    scope.update('a', dtype='integer', intent='inout')
    assert 'a' in scope.symbol_attrs
    assert scope.symbol_attrs['a'].dtype == INTEGER
    assert scope.symbol_attrs['a'].kind == 8  # Previous not wiped
    assert scope.symbol_attrs['a'].intent == 'inout'

    with pytest.raises(ValueError):
        scope.update('b', dtype='integer', intent='inout')

    # Override fail-safe, acts as another `declare()` call
    scope.update('b', dtype='integer', intent='inout', fail=False)
    assert 'b' in scope.symbol_attrs
    assert scope.symbol_attrs['b'].dtype == INTEGER
    assert scope.symbol_attrs['b'].intent == 'inout'


def test_scope_getter():
    """ Test basic :method:`get_type`/:method:`get_dtype` behaviour of :any:`Scope` """
    parent = Scope()
    scope = Scope(parent=parent)

    scope.declare('a', dtype='real', kind=8, intent='inout')
    parent.declare('b', dtype='integer', kind=4, intent='in')

    assert scope.get_type('a').dtype == REAL
    assert scope.get_type('a').kind == 8
    assert scope.get_type('a').intent == 'inout'

    # Non-recursive and recursive lookups through parent
    with pytest.raises(KeyError):
        scope.get_type('b', recursive=False)

    assert scope.get_type('b', recursive=False, fail=False) is None



# (C) Copyright 2018- ECMWF.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

"""
Verify correct frontend behaviour with regards to source parsing and sanitisation.
"""

import pytest

from loki import Module, Subroutine, Sourcefile, config_override
from loki.frontend import FP
from loki.ir import nodes as ir, FindNodes


@pytest.mark.parametrize('from_file', (True, False))
@pytest.mark.parametrize('preprocess', (True, False))
def test_source_sanitize_fp_source(tmp_path, from_file, preprocess):
    """
    Test that source sanitizing works as expected and postprocessing
    rules are correctly applied
    """
    fcode = """
subroutine some_routine(input_path)
    implicit none
    character(len=255), intent(in) :: input_path
    integer :: ios, fu
    write(*,*) "we print CPP value ", MY_VAR
    ! In the following line the PP definition should be replace by '0'
    ! or the actual line number
    write(*,*) "We are in line ",__LINE__
    open (action='read', file=TRIM(input_path), iostat=ios, newunit=fu)
end subroutine some_routine
""".strip()

    if from_file:
        filepath = tmp_path/'some_routine.F90'
        filepath.write_text(fcode)
        obj = Sourcefile.from_file(filepath, frontend=FP, preprocess=preprocess, defines=('MY_VAR=5',))
    else:
        obj = Sourcefile.from_source(fcode, frontend=FP, preprocess=preprocess, defines=('MY_VAR=5',))

    if preprocess:
        # CPP takes care of that
        assert '"We are in line ", 8' in obj.to_fortran()
        assert '"we print CPP value ", 5' in obj.to_fortran()
    else:
        # source sanitisation takes care of that
        assert '"We are in line ", 0' in obj.to_fortran()
        assert '"we print CPP value ", MY_VAR' in obj.to_fortran()

    assert 'newunit=fu' in obj.to_fortran()


@pytest.mark.parametrize('preprocess', (True, False))
def test_source_sanitize_fp_subroutine(preprocess):
    """
    Test that source sanitizing works as expected and postprocessing
    rules are correctly applied
    """
    fcode = """
subroutine some_routine(input_path)
    implicit none
    character(len=255), intent(in) :: input_path
    integer :: ios, fu
    write(*,*) "we print CPP value ", MY_VAR
    ! In the following line the PP definition should be replace by '0'
    ! or the actual line number
    write(*,*) "We are in line ",__LINE__
    open (action='read', file=TRIM(input_path), iostat=ios, newunit=fu)
end subroutine some_routine
""".strip()

    obj = Subroutine.from_source(fcode, frontend=FP, preprocess=preprocess, defines=('MY_VAR=5',))

    if preprocess:
        # CPP takes care of that
        assert '"We are in line ", 8' in obj.to_fortran()
        assert '"we print CPP value ", 5' in obj.to_fortran()
    else:
        # source sanitisation takes care of that
        assert '"We are in line ", 0' in obj.to_fortran()
        assert '"we print CPP value ", MY_VAR' in obj.to_fortran()

    assert 'newunit=fu' in obj.to_fortran()


@pytest.mark.parametrize('preprocess', (True, False))
def test_source_sanitize_fp_module(preprocess):
    """
    Test that source sanitizing works as expected and postprocessing
    rules are correctly applied
    """
    fcode = """
module some_mod
    implicit none
    integer line = __LINE__ + MY_VAR
contains
subroutine some_routine(input_path)
    implicit none
    character(len=255), intent(in) :: input_path
    integer :: ios, fu
    write(*,*) "we print CPP value ", MY_VAR
    ! In the following line the PP definition should be replace by '0'
    ! or the actual line number
    write(*,*) "We are in line ",__LINE__
    open (action='read', file=TRIM(input_path), iostat=ios, newunit=fu)
end subroutine some_routine
end module some_mod
""".strip()

    obj = Module.from_source(fcode, frontend=FP, preprocess=preprocess, defines=('MY_VAR=5',))

    if preprocess:
        # CPP takes care of that
        assert 'line = 3 + 5' in obj.to_fortran()
        assert '"We are in line ", 12' in obj.to_fortran()
        assert '"we print CPP value ", 5' in obj.to_fortran()
    else:
        # source sanitisation takes care of that
        assert 'line = 0 + MY_VAR' in obj.to_fortran()
        assert '"We are in line ", 0' in obj.to_fortran()
        assert '"we print CPP value ", MY_VAR' in obj.to_fortran()

    assert 'newunit=fu' in obj.to_fortran()


# TODO: Add tests for source sanitizer with other frontends


@pytest.mark.parametrize('store_source', (True, False))
def test_fparser_source_parsing(store_source):
    fcode = """
module test_source_mod
  use my_kind_mod, only: akind
  implicit none

  type my_type
    real(kind=akind) :: scalar, vector(3)
    integer :: asize
  end type my_type

contains

  subroutine my_test_routine(n, rick, dave)
    integer, intent(in) :: n
    real(kind=akind), intent(inout) :: rick, dave(n)
    integer :: i

    do i=1, n
      if (dave(i) > 0.5) then
        dave(i) = dave(i) + rick
      end if
    end do

    forall(i=1:n)
      dave(i) = dave(i) + 2.0
    end forall
  end subroutine my_test_routine
end module test_source_mod
"""
    with config_override({'frontend-store-source': store_source}):
        source = Sourcefile.from_source(fcode, frontend=FP)
        module = source['test_source_mod']
        routine = module['my_test_routine']

    if store_source:
        assert module.spec.source and module.spec.source.lines == (3, 10)
        assert module.contains.source and module.contains.source.lines == (11, 27)
        assert routine.spec.source and routine.spec.source.lines == (14, 16)
        assert routine.body.source and routine.body.source.lines == (17, 26)
    else:
        assert not module.spec.source
        assert not module.contains.source
        assert not routine.spec.source
        assert not routine.body.source

    decls = FindNodes(ir.VariableDeclaration).visit(routine.spec)
    loops = FindNodes(ir.Loop).visit(routine.body)
    conds = FindNodes(ir.Conditional).visit(routine.body)
    assigns = FindNodes(ir.Assignment).visit(routine.body)
    foralls = FindNodes(ir.Forall).visit(routine.body)
    assert len(decls) == 3 and len(loops) == 1 and len(conds) == 1
    assert len(assigns) == 2 and len(foralls) == 1

    if store_source:
        assert decls[0].source and decls[0].source.lines == (14, 14)
        assert decls[1].source and decls[1].source.lines == (15, 15)
        assert decls[2].source and decls[2].source.lines == (16, 16)
        assert loops[0].source and loops[0].source.lines == (18, 22)
        assert conds[0].source and conds[0].source.lines == (19, 21)
        assert assigns[0].source and assigns[0].source.lines == (20, 20)
        assert assigns[1].source and assigns[1].source.lines == (25, 25)
        assert foralls[0].source and foralls[0].source.lines == (24, 26)
    else:
        assert not decls[0].source and not decls[1].source and not decls[2].source
        assert not loops[0].source
        assert not conds[0].source
        assert not assigns[0].source

    imprts = FindNodes(ir.Import).visit(module.spec)
    intrs = FindNodes(ir.GenericStmt).visit(module.spec)
    tdefs = FindNodes(ir.TypeDef).visit(module.spec)
    assert len(imprts) == 1 and len(tdefs) == 1 and len(intrs) == 1
    tdecls = FindNodes(ir.VariableDeclaration).visit(tdefs[0].body)
    assert len(tdecls) == 2

    if store_source:
        assert imprts[0].source and imprts[0].source.lines == (3, 3)
        assert intrs[0].source and intrs[0].source.lines == (4, 4)
        assert tdefs[0].source and tdefs[0].source.lines == (6, 9)
        assert tdecls[0].source and tdecls[0].source.lines == (7, 7)
        assert tdecls[1].source and tdecls[1].source.lines == (8, 8)
    else:
        assert not imprts[0].source
        assert not intrs[0].source
        assert not tdefs[0].source
        assert not tdecls[0].source
        assert not tdecls[1].source


def test_fparser_sanitize_fypp_line_annotations():
    """
    Test that fypp line number annotations are sanitized correctly.
    """

    fcode = """
module some_templated_mod

# 1 "/path-to-hypp-macro/macro.hypp" 1
# 2 "/path-to-hypp-macro/macro.hypp"
# 3 "/path-to-hypp-macro/macro.hypp"
# 5 "/path-to-fypp-template/template.fypp" 2

integer :: a0
integer :: a1
integer :: a2
integer :: a3
integer :: a4

end module some_templated_mod
"""

    module = Module.from_source(fcode, frontend=FP)
    decls = FindNodes(ir.VariableDeclaration).visit(module.spec)

    assert len(decls) == 5



# (C) Copyright 2018- ECMWF.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

""" Intermediate node classes for nested node definitions in the Loki IR. """

from typing import Any, Tuple, Union, Optional

from pymbolic.primitives import Expression
from pydantic import field_validator

from loki.ir.nodes.abstract_nodes import (
    Node, InternalNode, ScopedNode
)
from loki.expression import (
    symbols as sym, AttachScopesMapper, ExpressionDimensionsMapper
)
from loki.tools import (
    as_tuple, dataclass_strict, flatten, sanitize_tuple, CaseInsensitiveDict
)
from loki.types import BasicType, SymbolAttributes


__all__ = [
    'Section', 'Associate', 'Loop', 'WhileLoop', 'Conditional',
    'PragmaRegion', 'Interface',
]


@dataclass_strict(frozen=True)
class _SectionBase():
    """ Type definitions for :any:`Section` node type. """


@dataclass_strict(frozen=True)
class Section(InternalNode, _SectionBase):
    """
    Internal representation of a single code region.
    """

    def __iter__(self):
        return iter(self.body)

    def __getitem__(self, index):
        return self.body[index]

    def __len__(self):
        return len(self.body)

    def __contains__(self, item):
        return item in self.body

    def index(self, item, start=0, stop=None):
        if stop is None:
            return self.body.index(item, start)
        return self.body.index(item, start, stop)

    def append(self, node):
        """
        Append the given node(s) to the section's body.

        Parameters
        ----------
        node : :any:`Node` or tuple of :any:`Node`
            The node(s) to append to the section.
        """
        self._update(body=self.body + as_tuple(node))

    def insert(self, pos, node):
        """
        Insert the given node(s) into the section's body at a specific
        position.

        Parameters
        ----------
        pos : int
            The position at which the node(s) should be inserted. Any existing
            nodes at this or after this position are shifted back.
        node : :any:`Node` or tuple of :any:`Node`
            The node(s) to append to the section.
        """
        self._update(body=self.body[:pos] + as_tuple(node) + self.body[pos:])  # pylint: disable=unsubscriptable-object

    def prepend(self, node):
        """
        Insert the given node(s) at the beginning of the section's body.

        Parameters
        ----------
        node : :any:`Node` or tuple of :any:`Node`
            The node(s) to insert into the section.
        """
        self._update(body=as_tuple(node) + self.body)

    def __repr__(self):
        if self.label is not None:
            return f'Section:: {self.label}'
        return 'Section::'


@dataclass_strict(frozen=True)
class _AssociateBase():
    """ Type definitions for :any:`Associate` node type. """

    associations: Tuple[Tuple[Expression, Expression], ...]


@dataclass_strict(frozen=True)
class Associate(ScopedNode, Section, _AssociateBase):  # pylint: disable=too-many-ancestors
    """
    Internal representation of a code region in which names are associated
    with expressions or variables.

    Parameters
    ----------
    body : tuple
        The associate's body.
    associations : dict or collections.OrderedDict
        The mapping of names to expressions or variables valid inside the
        associate's body.
    parent : :any:`Scope`, optional
        The parent scope in which the associate appears
    symbol_attrs : :any:`SymbolTable`, optional
        An existing symbol table to use
    **kwargs : optional
        Other parameters that are passed on to the parent class constructor.
    """

    _traversable = ['body', 'associations']

    def __bool__(self):
        return True

    def __post_init__(self, parent=None):
        super(ScopedNode, self).__post_init__(parent=parent)
        super(Section, self).__post_init__()

        assert self.associations is None or isinstance(self.associations, tuple)

    @property
    def association_map(self):
        """
        An :any:`collections.OrderedDict` of associated expressions.
        """
        return CaseInsensitiveDict((k, v) for k, v in self.associations)

    @property
    def inverse_map(self):
        """
        An :any:`collections.OrderedDict` of associated expressions.
        """
        return CaseInsensitiveDict((v, k) for k, v in self.associations)

    @property
    def variables(self):
        return tuple(v for _, v in self.associations)

    def _derive_local_symbol_types(self, parent_scope):
        """ Derive the types of locally defined symbols from their associations. """

        rescoped_associations = ()
        for expr, name in self.associations:
            # Put symbols in associated expression into the right scope
            expr = AttachScopesMapper()(expr, scope=parent_scope)

            # Determine type of new names
            if isinstance(expr, (sym.TypedSymbol, sym.MetaSymbol)):
                # Use the type of the associated variable
                _type = expr.type.clone(parent=None)
                if isinstance(expr, sym.Array) and expr.dimensions is not None:
                    shape = ExpressionDimensionsMapper()(expr)
                    if shape == (sym.IntLiteral(1),):
                        # For a scalar expression, we remove the shape
                        shape = None
                    _type = _type.clone(shape=shape)
            else:
                # TODO: Handle data type and shape of complex expressions
                shape = ExpressionDimensionsMapper()(expr)
                if shape == (sym.IntLiteral(1),):
                    # For a scalar expression, we remove the shape
                    shape = None
                _type = SymbolAttributes(BasicType.DEFERRED, shape=shape)
            name = name.clone(scope=self, type=_type)
            rescoped_associations += ((expr, name),)

        self._update(associations=rescoped_associations)

    def __repr__(self):
        if self.associations:
            associations = ', '.join(f'{str(var)}={str(expr)}'
                                     for var, expr in self.associations)
            return f'Associate:: {associations}'
        return 'Associate::'


@dataclass_strict(frozen=True)
class _LoopBase():
    """ Type definitions for :any:`Loop` node type. """

    variable: Expression
    bounds: Expression
    body: Tuple[Node, ...]
    pragma: Optional[Tuple[Node, ...]] = None
    pragma_post: Optional[Tuple[Node, ...]] = None
    loop_label: Optional[Any] = None
    name: Optional[str] = None
    has_end_do: Optional[bool] = True


@dataclass_strict(frozen=True)
class Loop(InternalNode, _LoopBase):
    """
    Internal representation of a loop with induction variable and range.

    Parameters
    ----------
    variable : :any:`Scalar`
        The induction variable of the loop.
    bounds : :any:`LoopRange`
        The range of the loop, defining the iteration space.
    body : tuple
        The loop body.
    pragma : tuple of :any:`Pragma`, optional
        Pragma(s) that appear in front of the loop. By default :any:`Pragma`
        nodes appear as standalone nodes in the IR before the :any:`Loop` node.
        Only a bespoke context created by :py:func:`pragmas_attached`
        attaches them for convenience.
    pragma_post : tuple of :any:`Pragma`, optional
        Pragma(s) that appear after the loop. The same applies as for `pragma`.
    loop_label : str, optional
        The Fortran label for that loop. Importantly, this is an intrinsic
        Fortran feature and different from the statement label that can be
        attached to other nodes.
    name : str, optional
        The Fortran construct name for that loop.
    has_end_do : bool, optional
        In Fortran, loop blocks can be closed off by a ``CONTINUE`` statement
        (which we retain as an :any:`Intrinsic` node) and therefore ``END DO``
        can be omitted. For string reproducibility this parameter can be set
        `False` to indicate that this loop did not have an ``END DO``
        statement in the original source.
    **kwargs : optional
        Other parameters that are passed on to the parent class constructor.
    """

    _traversable = ['variable', 'bounds', 'body']

    def __post_init__(self):
        super().__post_init__()
        assert self.variable is not None

    def __repr__(self):
        label = ', '.join(l for l in [self.name, self.loop_label] if l is not None)
        if label:
            label = ' ' + label
        control = f'{str(self.variable)}={str(self.bounds)}'
        return f'Loop::{label} {control}'


@dataclass_strict(frozen=True)
class _WhileLoopBase():
    """ Type definitions for :any:`WhileLoop` node type. """

    condition: Optional[Expression]
    body: Tuple[Node, ...]
    pragma: Optional[Node] = None
    pragma_post: Optional[Node] = None
    loop_label: Optional[Any] = None
    name: Optional[str] = None
    has_end_do: Optional[bool] = True


@dataclass_strict(frozen=True)
class WhileLoop(InternalNode, _WhileLoopBase):
    """
    Internal representation of a while loop in source code.

    Importantly, this is different from a ``DO`` (Fortran) or ``for`` (C) loop,
    as we do not have a specified induction variable with explicit iteration
    range.

    Parameters
    ----------
    condition : :any:`pymbolic.primitives.Expression`
        The condition evaluated before executing the loop body.
    body : tuple
        The loop body.
    pragma : tuple of :any:`Pragma`, optional
        Pragma(s) that appear in front of the loop. By default :any:`Pragma`
        nodes appear as standalone nodes in the IR before the :any:`Loop` node.
        Only a bespoke context created by :py:func:`pragmas_attached`
        attaches them for convenience.
    pragma_post : tuple of :any:`Pragma`, optional
        Pragma(s) that appear after the loop. The same applies as for `pragma`.
    loop_label : str, optional
        The Fortran label for that loop. Importantly, this is an intrinsic
        Fortran feature and different from the statement label that can be
        attached to other nodes.
    name : str, optional
        The Fortran construct name for that loop.
    has_end_do : bool, optional
        In Fortran, loop blocks can be closed off by a ``CONTINUE`` statement
        (which we retain as an :any:`Intrinsic` node) and therefore ``END DO``
        can be omitted. For string reproducibility this parameter can be set
        `False` to indicate that this loop did not have an ``END DO``
        statement in the original source.
    **kwargs : optional
        Other parameters that are passed on to the parent class constructor.
    """

    _traversable = ['condition', 'body']

    def __repr__(self):
        label = ', '.join(l for l in [self.name, self.loop_label] if l is not None)
        if label:
            label = ' ' + label
        control = str(self.condition) if self.condition else ''
        return f'WhileLoop::{label} {control}'


@dataclass_strict(frozen=True)
class _ConditionalBase():
    """ Type definitions for :any:`Conditional` node type. """

    condition: Expression
    body: Tuple[Node, ...]
    else_body: Optional[Tuple[Node, ...]] = ()
    inline: bool = False
    has_elseif: bool = False
    name: Optional[str] = None


@dataclass_strict(frozen=True)
class Conditional(InternalNode, _ConditionalBase):
    """
    Internal representation of a conditional branching construct.

    Parameters
    ----------
    condition : :any:`pymbolic.primitives.Expression`
        The condition evaluated before executing the body.
    body : tuple
        The conditional's body.
    else_body : tuple
        The body of the else branch. Can be empty.
    inline : bool, optional
        Flag that marks this conditional as inline, i.e., it s body consists
        only of a single statement that appeared immediately after the
        ``IF`` statement and it does not have an ``else_body``.
    has_elseif : bool, optional
        Flag that indicates that this conditional has an ``ELSE IF`` branch
        in the original source. In Loki's IR these are represented as a chain
        of :any:`Conditional` but for string reproducibility this flag can be
        provided to enable backends to reproduce the original appearance.
    name : str, optional
        The Fortran construct name for that conditional.
    **kwargs : optional
        Other parameters that are passed on to the parent class constructor.
    """

    _traversable = ['condition', 'body', 'else_body']

    @field_validator('body', 'else_body', mode='before')
    @classmethod
    def ensure_tuple(cls, value):
        return sanitize_tuple(value)

    def __post_init__(self):
        super().__post_init__()
        assert self.condition is not None

        if self.has_elseif:
            assert len(self.else_body) == 1
            assert isinstance(self.else_body[0], Conditional)  # pylint: disable=unsubscriptable-object

    def __repr__(self):
        if self.name:
            return f'Conditional:: {self.name}'
        return 'Conditional::'

    @property
    def else_bodies(self):
        """
        Return all nested node tuples in the ``ELSEIF``/``ELSE`` part
        of the conditional chain.
        """
        if self.has_elseif:
            return (self.else_body[0].body,) + self.else_body[0].else_bodies
        return (self.else_body,) if self.else_body else ()


@dataclass_strict(frozen=True)
class _PragmaRegionBase():
    """ Type definitions for :any:`PragmaRegion` node type. """

    body: Tuple[Node, ...]
    pragma: Node = None
    pragma_post: Node = None


@dataclass_strict(frozen=True)
class PragmaRegion(InternalNode, _PragmaRegionBase):
    """
    Internal representation of a block of code defined by two matching pragmas.

    Generally, the pair of pragmas are assumed to be of the form
    ``!$<keyword> <marker>`` and ``!$<keyword> end <marker>``.

    This node type is injected into the IR within a context created by
    :py:func:`pragma_regions_attached`.

    Parameters
    ----------
    body : tuple
        The statements appearing between opening and closing pragma.
    pragma : :any:`Pragma`
        The opening pragma declaring that region.
    pragma_post : :any:`Pragma`
        The closing pragma for that region.
    **kwargs : optional
        Other parameters that are passed on to the parent class constructor.
    """

    _traversable = ['body']

    def append(self, node):
        self._update(body=self.body + as_tuple(node))

    def insert(self, pos, node):
        '''Insert at given position'''
        self._update(body=self.body[:pos] + as_tuple(node) + self.body[pos:])  # pylint: disable=unsubscriptable-object

    def prepend(self, node):
        self._update(body=as_tuple(node) + self.body)

    def __repr__(self):
        return 'PragmaRegion::'


@dataclass_strict(frozen=True)
class _InterfaceBase():
    """ Type definitions for :any:`Interface` node type. """

    body: Tuple[Any, ...]
    abstract: bool = False
    spec: Optional[Union[Expression, str]] = None


@dataclass_strict(frozen=True)
class Interface(InternalNode, _InterfaceBase):
    """
    Internal representation of a Fortran interface block.

    Parameters
    ----------
    body : tuple
        The body of the interface block, containing function and subroutine
        specifications or procedure statements
    abstract : bool, optional
        Flag to indicate that this is an abstract interface
    spec : str, optional
        A generic name, operator, assignment, or I/O specification
    **kwargs : optional
        Other parameters that are passed on to the parent class constructor.
    """

    _traversable = ['body']

    def __post_init__(self):
        super().__post_init__()
        assert not (self.abstract and self.spec)

    @property
    def symbols(self):
        """
        The list of symbol names declared by this interface
        """
        symbols = as_tuple(flatten(
            getattr(node, 'procedure_symbol', getattr(node, 'symbols', ()))
            for node in self.body  # pylint: disable=not-an-iterable
        ))
        if self.spec:
            return (self.spec,) + symbols
        return symbols

    @property
    def symbol_map(self):
        """
        Map symbol name to symbol declared by this interface
        """
        return CaseInsensitiveDict(
            (s.name.lower(), s) for s in self.symbols
        )

    def __contains__(self, name):
        return name in self.symbol_map

    def __repr__(self):
        symbols = ', '.join(str(var) for var in self.symbols)
        if self.abstract:
            return f'Abstract Interface:: {symbols}'
        if self.spec:
            return f'Interface {self.spec}:: {symbols}'
        return f'Interface:: {symbols}'



# (C) Copyright 2018- ECMWF.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

"""
Verify correct frontend behaviour for source-level model handling.
"""

from pathlib import Path

import pytest

from loki import (
    Module, Sourcefile, PreprocessorDirective, Assignment,
    Comment, CommentBlock, RawSource, Scalar, config, config_override
)
from loki.frontend import available_frontends, OMNI, FP, REGEX
from loki.ir import nodes as ir, FindNodes


@pytest.fixture(scope='module', name='here')
def fixture_here():
    return Path(__file__).parents[2] / 'tests'


@pytest.mark.parametrize('frontend', available_frontends())
@pytest.mark.usefixtures('reset_frontend_mode')
def test_frontend_strict_mode(frontend, tmp_path):
    """
    Verify that frontends fail on unsupported features if strict mode is enabled
    """
    # Parameterized derived types currently not implemented
    fcode = """
module frontend_strict_mode
    implicit none
    TYPE matrix ( k, b )
      INTEGER,     KIND :: k = 4
      INTEGER (8), LEN  :: b
      REAL (k)          :: element (b,b)
    END TYPE matrix
end module frontend_strict_mode
    """
    config['frontend-strict-mode'] = True
    with pytest.raises(NotImplementedError):
        Module.from_source(fcode, frontend=frontend, xmods=[tmp_path])

    config['frontend-strict-mode'] = False
    module = Module.from_source(fcode, frontend=frontend, xmods=[tmp_path])
    assert 'matrix' in module.symbol_attrs
    assert 'matrix' in module.typedef_map


@pytest.mark.parametrize('frontend', available_frontends())
def test_frontend_main_program(frontend):
    """
    Loki can't handle PROGRAM blocks and the frontends should throw an exception
    """
    fcode = """
program hello
    print *, "Hello World!"
end program
    """.strip()

    with config_override({'frontend-strict-mode': True}):
        with pytest.raises(NotImplementedError):
            Sourcefile.from_source(fcode, frontend=frontend)

    source = Sourcefile.from_source(fcode, frontend=frontend)
    assert source.ir.body == ()


@pytest.mark.parametrize('frontend', available_frontends())
def test_frontend_source_lineno(frontend):
    """
    ...
    """
    fcode = """
    subroutine driver
        call kernel()
        call kernel()
        call kernel()
    end subroutine driver
    """

    source = Sourcefile.from_source(fcode, frontend=frontend)
    routine = source['driver']
    calls = FindNodes(ir.CallStatement).visit(routine.body)
    assert calls[0] != calls[1]
    assert calls[1] != calls[2]
    assert calls[0].source.lines[0] < calls[1].source.lines[0] < calls[2].source.lines[0]


@pytest.mark.parametrize('frontend', available_frontends(
    include_regex=True, skip=[(OMNI, 'OMNI may segfault on empty files')]
))
@pytest.mark.parametrize('fcode', ['', '\n', '\n\n\n\n'])
def test_frontend_empty_file(frontend, fcode):
    """Ensure that all frontends can handle empty source files correctly (#186)"""
    source = Sourcefile.from_source(fcode, frontend=frontend)
    assert isinstance(source.ir, ir.Section)
    assert not source.to_fortran().strip()


@pytest.mark.parametrize('frontend', available_frontends())
def test_sourcefile_properties(here, frontend, tmp_path):
    """
    Test that all subroutines and functions are discovered
    and exposed via `subroutines` and `all_subroutines` properties.
    """
    filepath = here/'sources/sourcefile.f90'
    source = Sourcefile.from_file(filepath, frontend=frontend, xmods=[tmp_path])
    assert len(source.subroutines) == 3
    assert len(source.all_subroutines) == 5

    subroutines = ['routine_a', 'routine_b', 'function_d']
    all_subroutines = subroutines + ['module_routine', 'module_function']
    contained_routines = ['contained_c']

    assert sum(routine.name in subroutines for routine in source.subroutines) == 3
    assert sum(routine.name in all_subroutines for routine in source.subroutines) == 3
    assert sum(routine.name in contained_routines for routine in source.subroutines) == 0

    assert sum(routine.name in subroutines for routine in source.all_subroutines) == 3
    assert sum(routine.name in all_subroutines for routine in source.all_subroutines) == 5
    assert sum(routine.name in contained_routines for routine in source.all_subroutines) == 0


@pytest.mark.parametrize('frontend', available_frontends())
def test_sourcefile_from_source(frontend, tmp_path):
    """
    Test the `from_source` constructor for `Sourcefile` objects.
    """

    fcode = """
! Some comment
subroutine routine_a
  integer a
  a = 1
end subroutine routine_a

! Some comment
module some_module
contains
  subroutine module_routine
    integer m
    m = 2
  end subroutine module_routine
  function module_function(n)
    integer n
    n = 3
  end function module_function
end module some_module
! Other comment

subroutine routine_b
  integer b
  b = 4
contains
  subroutine contained_c
    integer c
    c = 5
  end subroutine contained_c
end subroutine routine_b
! Other comment

function function_d(d)
  integer d
  d = 6
end function function_d
""".strip()
    source = Sourcefile.from_source(fcode, frontend=frontend, xmods=[tmp_path])
    assert len(source.subroutines) == 3
    assert len(source.all_subroutines) == 5

    subroutines = ['routine_a', 'routine_b', 'function_d']
    all_subroutines = subroutines + ['module_routine', 'module_function']

    assert [routine.name.lower() for routine in source.subroutines] == subroutines
    assert [routine.name.lower() for routine in source.all_subroutines] == all_subroutines
    assert 'contained_c' not in [routine.name.lower() for routine in source.subroutines]
    assert 'contained_c' not in [routine.name.lower() for routine in source.all_subroutines]

    comments = FindNodes((Comment, CommentBlock)).visit(source.ir)
    assert len(comments) == 4
    assert all(comment.text.strip() in ['! Some comment', '! Other comment'] for comment in comments)


@pytest.mark.parametrize('frontend', available_frontends(
    skip={OMNI, 'F_Front segfault is a known issue on some platforms'}
))
def test_sourcefile_lazy_construction(frontend, tmp_path):
    """
    Test delayed ("lazy") parsing of sourcefile content
    """
    fcode = """
! A comment to test
subroutine routine_a
integer a
a = 1
end subroutine routine_a

module some_module
contains
subroutine module_routine
integer m
m = 2
end subroutine module_routine
function module_function(n)
integer n
n = 3
end function module_function
end module some_module

#ifndef SOME_PREPROC_VAR
subroutine routine_b
integer b
b = 4
contains
subroutine contained_c
integer c
c = 5
end subroutine contained_c
end subroutine routine_b
#endif

function function_d(d)
integer d
d = 6
end function function_d
    """.strip()
    source = Sourcefile.from_source(fcode, frontend=REGEX)
    assert len(source.subroutines) == 3
    assert len(source.all_subroutines) == 5

    some_module = source['some_module']
    routine_b = source['routine_b']
    module_routine = some_module['module_routine']
    function_d = source['function_d']
    assert function_d.arguments == ()

    # Make sure we have an incomplete parse tree until now
    assert source._incomplete
    assert len(FindNodes(RawSource).visit(source.ir)) == 5
    assert len(FindNodes(RawSource).visit(source['routine_a'].ir)) == 1

    # Trigger the full parse
    source.make_complete(frontend=frontend, xmods=[tmp_path])
    assert not source._incomplete

    # Make sure no RawSource nodes are left
    assert not FindNodes(RawSource).visit(source.ir)
    if frontend == FP:
        # Some newlines are also treated as comments
        assert len(FindNodes(Comment).visit(source.ir)) == 2
    else:
        assert len(FindNodes(Comment).visit(source.ir)) == 1
    if frontend == OMNI:
        assert not FindNodes(PreprocessorDirective).visit(source.ir)
    else:
        assert len(FindNodes(PreprocessorDirective).visit(source.ir)) == 2
    for routine in source.all_subroutines:
        assert not FindNodes(RawSource).visit(routine.ir)
        assert len(FindNodes(Assignment).visit(routine.ir)) == 1

    # The previously generated ProgramUnit objects should be the same as before
    assert routine_b is source['routine_b']
    assert some_module is source['some_module']
    assert module_routine is source['some_module']['module_routine']
    assert function_d.arguments == ('d',)
    assert isinstance(function_d.arguments[0], Scalar)


@pytest.mark.parametrize('frontend', available_frontends())
def test_sourcefile_lazy_comments(frontend):
    """
    Make sure that lazy construction can handle comments on source file level
    (i.e. outside a program unit)
    """
    fcode = """
! Comment outside
subroutine myroutine
    ! Comment inside
end subroutine myroutine
! Other comment outside
    """.strip()
    source = Sourcefile.from_source(fcode, frontend=REGEX)

    assert isinstance(source.ir.body[0], RawSource)
    assert isinstance(source.ir.body[2], RawSource)

    myroutine = source['myroutine']
    assert isinstance(myroutine.spec[0], RawSource)

    source.make_complete(frontend=frontend)

    assert isinstance(source.ir.body[0], Comment)
    assert isinstance(source.ir.body[2], Comment)
    if frontend == OMNI:
        assert isinstance(myroutine.body.body[0], Comment)
    else:
        assert isinstance(myroutine.docstring[0], Comment)

    code = source.to_fortran()
    assert '! Comment outside' in code
    assert '! Comment inside' in code
    assert '! Other comment outside' in code



# (C) Copyright 2024- ECMWF.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from copy import deepcopy

import itertools

from loki.expression import (
    symbols as sym, get_pyrange, is_constant, SimplifyMapper
)
from loki.ir import nodes as ir, FindNodes, FindVariables, Transformer
from loki.subroutine import Subroutine
from loki.tools import dict_override

from loki.transformations.transform_loop import LoopUnrollTransformer


__all__ = [
    'do_constant_propagation', 'ConstantPropagationMapper',
    'ConstantPropagationTransformer'
]


def _pop_array_accesses(lhs, **kwargs):
    constants_map = kwargs.get('constants_map', {})
    new_shape = ConstantPropagationMapper()(lhs.shape, constants_map=constants_map)

    literal_mask = [is_constant(dimension) for dimension in lhs.dimensions]
    computable_dimension_mask = [is_constant(extent) for extent in new_shape]

    masked_indices = []
    for literal, computable, dimension in zip(literal_mask, computable_dimension_mask, lhs.dimensions):
        if literal:
            masked_indices.append(dimension)
        elif computable:
            masked_indices.append(sym.RangeIndex((None, None, None)))
        else:
            masked_indices.append(-1)

    possible_accesses = _array_indices_to_accesses(masked_indices, new_shape)
    for access in possible_accesses:
        constants_map.pop((lhs.basename, access), None)


def update_constants_map(lhs, value, constants_map):
    constants_map[(lhs.basename, ())] = value


def invalidate_constants_map(lhs, constants_map):
    if isinstance(lhs, sym.Array):
        for access in tuple(key for key in constants_map if key[0] == lhs.basename):
            constants_map.pop((lhs.basename, access), None)
        return

    constants_map.pop((lhs.basename, ()), None)


def _separate_literals(children):
    separated = ([], [])
    for child in children:
        if isinstance(child, sym._Literal):
            separated[0].append(child)
        else:
            separated[1].append(child)
    return separated


def _array_indices_to_accesses(dimensions, shape):
    arg_lists = []
    for count, dimension in enumerate(dimensions):
        if isinstance(dimension, sym.RangeIndex):
            start = dimension.start if dimension.start else sym.IntLiteral(1)
            stop = dimension.stop if dimension.stop else shape[count]
            arg_lists.append([
                sym.IntLiteral(v) for v in get_pyrange(sym.LoopRange((start, stop, dimension.step)))
            ])
        else:
            arg_lists.append([dimension])
    return itertools.product(*arg_lists)


class ConstantPropagationMapper(SimplifyMapper):
    """ Mapper for expression-level constant replacement and folding. """

    def map_array(self, expr, *args, **kwargs):
        constants_map = kwargs.get('constants_map', {})
        return constants_map.get((expr.basename, getattr(expr, 'dimensions', ())), expr)

    def map_quotient(self, expr, *args, **kwargs):
        """ Always force-evaluate integer-division """
        if isinstance(expr.numerator, sym.IntLiteral) and isinstance(expr.denominator, sym.IntLiteral):
            return sym.IntLiteral(float(expr.numerator.value) / float(expr.denominator.value))
        return super().map_quotient(expr, *args, **kwargs)

    map_scalar = map_array
    map_deferred_type_symbol = map_array


class ConstantPropagationTransformer(Transformer):
    """Apply constant-propagation analysis as a transformation driver."""

    def visit_Assignment(self, o, **kwargs):
        constants_map = kwargs.get('constants_map', {})
        mapper = ConstantPropagationMapper()

        rhs_symbols = FindVariables().visit(o.rhs)
        if kwargs.get('within_loop', False) and o.lhs in rhs_symbols:
            # In loop bodies, skip "increment" updates to the LHS value
            return o

        # Resolve known constants on the RHS
        new_rhs = mapper(o.rhs, constants_map=constants_map)
        new_lhs = o.lhs

        if isinstance(o.lhs, sym.Array):
            new_dimensions = tuple(mapper(d, constants_map=constants_map) for d in o.lhs.dimensions)
            new_lhs = o.lhs.clone(dimensions=new_dimensions)

            _, non_literal_dimensions = _separate_literals(new_dimensions)
            if non_literal_dimensions:
                _pop_array_accesses(new_lhs, constants_map=constants_map)
                return o._rebuild(lhs=new_lhs, rhs=new_rhs)

        _, non_literals = _separate_literals((new_rhs,))
        if not non_literals and not isinstance(new_lhs, sym.Array):
            update_constants_map(new_lhs, new_rhs, constants_map)
        else:
            invalidate_constants_map(new_lhs, constants_map)

        return o._rebuild(lhs=new_lhs, rhs=new_rhs)

    def visit_Conditional(self, o, **kwargs):
        constants_map = kwargs.get('constants_map', {})
        mapper = ConstantPropagationMapper()

        new_condition = mapper(o.condition, constants_map=constants_map)

        # Pass two copies of the constants map forward ...
        with dict_override(kwargs, {'constants_map': deepcopy(constants_map)}):
            new_body = self.visit(o.body, **kwargs)
            body_constants_map = kwargs['constants_map']
        with dict_override(kwargs, {'constants_map': deepcopy(constants_map)}):
            new_else_body = self.visit(o.else_body, **kwargs)
            else_constants_map = kwargs['constants_map']

        # ... then merge the maps, removing all non-shared entries
        merged_constants_map = {}
        all_keys = set(body_constants_map) | set(else_constants_map)
        for key in all_keys:
            if (
                    key in body_constants_map and key in else_constants_map
                    and body_constants_map[key] == else_constants_map[key]
            ):
                merged_constants_map[key] = body_constants_map[key]
            else:
                merged_constants_map.pop(key, None)

        # Update the shared constants map with the merged result
        constants_map.clear()
        constants_map.update(merged_constants_map)

        return o._rebuild(condition=new_condition, body=new_body, else_body=new_else_body)

    def visit_Loop(self, o, **kwargs):
        constants_map = kwargs.get('constants_map', {})
        mapper = ConstantPropagationMapper()

        new_bounds = mapper(o.bounds, constants_map=constants_map)

        # When recursing into loops, send a flag down to trigger detection
        # of loop-variant assignments ("increment" updates to variables).
        with dict_override(kwargs, {
                'within_loop': True, 'constants_map': deepcopy(constants_map)
        }):
            kwargs['constants_map'].pop((o.variable.basename, ()), None)
            new_body = self.visit(o.body, **kwargs)

        lhs_vars = {o.variable}
        lhs_vars.update(loop.variable for loop in FindNodes(ir.Loop).visit(o.body))

        assignments = FindNodes(ir.Assignment).visit(new_body)
        for assign in assignments:
            lhs_vars.add(assign.lhs)

        bounds_are_const = (
            is_constant(new_bounds.start)
            and is_constant(new_bounds.stop)
            and (is_constant(new_bounds.step) or new_bounds.step is None)
        )

        if bounds_are_const:
            loop_constants_map = constants_map

            for assign in assignments:
                if not set(FindVariables().visit(assign.rhs)).intersection(lhs_vars):
                    assign_kwargs = dict(kwargs)
                    assign_kwargs['constants_map'] = loop_constants_map
                    self.visit_Assignment(assign, **assign_kwargs)
        else:
            for assign in assignments:
                invalidate_constants_map(assign.lhs, constants_map)

        invalidate_constants_map(o.variable, constants_map)

        return o._rebuild(bounds=new_bounds, body=new_body)

    def generate_declarations_map(self, routine):
        """Build the initial constant map from declaration-time initializers."""

        def index_initial_elements(indices, element):
            if len(indices) == 1:
                return element.elements[indices[0].value - 1]
            return index_initial_elements(indices[1:], element.elements[indices[0].value - 1])

        declarations_map = {}
        for symbol in getattr(routine, 'symbols', ()):
            if isinstance(symbol, sym.DeferredTypeSymbol) or symbol.initial is None:
                continue

            if isinstance(symbol, sym.Array):
                declarations_map.update({
                    (symbol.basename, indices): index_initial_elements(indices, symbol.initial)
                    for indices in _array_indices_to_accesses(
                        [sym.RangeIndex((None, None, None))] * len(symbol.shape), symbol.shape
                    )
                })
            else:
                declarations_map[(symbol.basename, ())] = symbol.initial
        return declarations_map


def do_constant_propagation(routine, unroll_loops=False):
    """ Apply constant-propagation over the body of a :any:`Subroutine`. """

    assert isinstance(routine, Subroutine), \
        f'[Loki] Constant propagation can only be applied to Subroutine, but found {routine}'

    const_prop = ConstantPropagationTransformer(inplace=True, invalidate_source=False)
    declarations_map = const_prop.generate_declarations_map(routine)

    if routine.spec:
        routine.spec = const_prop.visit(routine.spec, constants_map=declarations_map)
    if routine.body:
        routine.body = const_prop.visit(routine.body, constants_map=declarations_map)

    if unroll_loops:
        routine.body = LoopUnrollTransformer().visit(routine.body)

        # If loop unrolling is requested, do another forward propagation pass
        routine.body = const_prop.visit(routine.body, constants_map=declarations_map)

    return routine



# (C) Copyright 2018- ECMWF.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

"""
Verify correct frontend behaviour for control-flow IR nodes.
"""

import pytest

from loki import Module, Subroutine
from loki.frontend import available_frontends, OMNI
from loki.ir import nodes as ir, FindNodes


@pytest.mark.parametrize('frontend', available_frontends(skip=[(OMNI, 'OMNI fails to read without full module')]))
def test_select_type(frontend, tmp_path):
    fcode = """
module select_type_mod
    use imported_type_mod, only: imported_type
    implicit none
    type, abstract :: base
    end type base
    type, extends(base) :: derived1
        real :: val
    end type derived1
    type, extends(base) :: derived2
        integer :: val
    end type derived2
contains
    subroutine select_type_routine(arg, arg2)
        class(base), intent(inout) :: arg
        class(imported_type), intent(inout) :: arg2
        select type( arg )
            class is(derived1)
                arg%val = 1.0
            class is(derived2)
                arg%val = 1
            class default
                print *, 'error'
        end select
        ! Some comment before the second select
        select type( arg )
            type is(base)
                write(*,*) 'default'
        end select
        select type( arg2 )
            ! inline comment
            type is(imported_type)
                print *, 'imported type'
        end select
    end subroutine select_type_routine
end module select_type_mod
    """.strip()

    module = Module.from_source(fcode, frontend=frontend, xmods=[tmp_path])
    tconds = FindNodes(ir.TypeConditional).visit(module['select_type_routine'].body)
    assert len(tconds) == 3

    assert tconds[0].expr == 'arg'
    assert tconds[0].values == (
        ('derived1', True), ('derived2', True)
    )
    assert len(tconds[0].bodies) == 2
    assert len(tconds[0].else_body) == 1

    assert tconds[1].expr == 'arg'
    assert tconds[1].values == (('base', False),)
    assert not tconds[1].else_body

    assert tconds[2].expr == 'arg2'
    assert tconds[2].values == (('imported_type', False),)
    assert not tconds[2].else_body

    comments = FindNodes(ir.Comment).visit(module['select_type_routine'].body)
    assert len(comments) == 2
    assert 'Some comment' in comments[0].text
    assert 'inline comment' in comments[1].text


@pytest.mark.parametrize('frontend', available_frontends())
def test_select_case(frontend):
    """
    Test `SELECT CASE` parsing into multi-conditional IR.
    """
    fcode = """
subroutine select_case_routine(kind, out)
  implicit none
  integer, intent(in) :: kind
  integer, intent(out) :: out

  select case (kind)
  case (1)
    out = 10
  case (2:4)
    out = 20
  case default
    out = 30
  end select
end subroutine select_case_routine
    """.strip()

    routine = Subroutine.from_source(fcode, frontend=frontend)
    multi_conds = FindNodes(ir.MultiConditional).visit(routine.body)
    assert len(multi_conds) == 1
    assert multi_conds[0].expr == 'kind'
    assert len(multi_conds[0].values) == 2
    assert len(multi_conds[0].bodies) == 2
    assert len(multi_conds[0].else_body) == 1


@pytest.mark.parametrize('frontend', available_frontends())
def test_do_while_loop(frontend):
    """
    Test `DO WHILE` parsing.
    """
    fcode = """
subroutine do_while_routine(n, out)
  implicit none
  integer, intent(in) :: n
  integer, intent(out) :: out
  integer :: i

  out = 0
  i = 1
  do while (i <= n)
    out = out + i
    i = i + 1
  end do
end subroutine do_while_routine
    """.strip()

    routine = Subroutine.from_source(fcode, frontend=frontend)
    whiles = FindNodes(ir.WhileLoop).visit(routine.body)
    assert len(whiles) == 1
    assert whiles[0].condition == 'i <= n'
    assigns = FindNodes(ir.Assignment).visit(whiles[0].body)
    assert len(assigns) == 2


@pytest.mark.parametrize('frontend', available_frontends())
def test_control_flow_statements(frontend):
    """
    Test Fortran statements that define control flow in loop bodies.
    """
    fcode = """
subroutine control_flow_routine(n, out)
  implicit none
  integer, intent(in) :: n
  integer, intent(out) :: out
  integer :: i

  out = 0
  do i=1, n
    if (i == 2) cycle
    if (i == 3) continue
    if (i == 4) go to 42
    if (i == 5) then
42    exit
    end if
    if (i == 6) stop
    out = out + i
  end do
end subroutine control_flow_routine
    """.strip()

    routine = Subroutine.from_source(fcode, frontend=frontend)
    loops = FindNodes(ir.Loop).visit(routine.body)
    assert len(loops) == 1
    statements = FindNodes(ir.GenericStmt).visit(loops[0].body)

    assert len(statements) == 5
    assert isinstance(statements[0], ir.CycleStmt)
    assert isinstance(statements[1], ir.ContinueStmt)
    assert isinstance(statements[2], ir.GotoStmt)
    assert statements[2].text == '42'
    assert isinstance(statements[3], ir.ExitStmt)
    assert isinstance(statements[4], ir.StopStmt)


@pytest.mark.parametrize('frontend', available_frontends())
def test_where_construct(frontend):
    """
    Test `WHERE` handling for frontends that lower it to intrinsic/raw bodies.
    """
    fcode = """
subroutine where_routine(mask, a, b)
  implicit none
  logical, intent(in) :: mask(:)
  real, intent(inout) :: a(:), b(:)

  where (mask)
    a = b
  elsewhere
    a = 0.
  end where
end subroutine where_routine
    """.strip()

    routine = Subroutine.from_source(fcode, frontend=frontend)
    masked = FindNodes(ir.MaskedStatement).visit(routine.body)
    assert len(masked) == 1
    assert len(masked[0].conditions) == 1
    assert masked[0].conditions[0] == 'mask'
    assert len(masked[0].bodies) == 1
    assert len(masked[0].default) == 1



# (C) Copyright 2018- ECMWF.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

"""
Verify correct frontend behaviour for procedure declarations and interfaces.
"""

from pathlib import Path

import pytest

from loki import Sourcefile, Subroutine, fgen
from loki.expression import symbols as sym
from loki.frontend import available_frontends, OMNI
from loki.ir import nodes as ir, FindNodes
from loki.jit_build import jit_compile, jit_compile_lib, clean_test
from loki.types import BasicType, ProcedureType, SymbolAttributes


@pytest.fixture(scope='module', name='header_path')
def fixture_header_path():
    return Path(__file__).parents[2] / 'tests' / 'sources' / 'header.f90'


@pytest.mark.parametrize('frontend', available_frontends())
def test_external_stmt(tmp_path, frontend):
    """
    Tests procedures passed as dummy arguments and declared as EXTERNAL.
    """
    fcode_external = """
! This should be tested as well with interface statements in the caller
! routine, and the subprogram definitions outside (to have "truly external"
! procedures, however, we need to make the INTERFACE support more robust first

subroutine other_external_subroutine(outvar)
  implicit none
  integer, intent(out) :: outvar
  outvar = 4
end subroutine other_external_subroutine

function other_external_function() result(outvar)
  implicit none
  integer :: outvar
  outvar = 6
end function other_external_function
    """.strip()

    fcode = """
subroutine routine_external_stmt(invar, sub1, sub2, sub3, outvar, func1, func2, func3)
  implicit none
  integer, intent(in) :: invar
  external sub1
  external :: sub2, sub3
  integer, intent(out) :: outvar
  integer, external :: func1, func2
  integer, external :: func3
  integer tmp

  call sub1(tmp)
  outvar = invar + tmp  ! invar + 1
  call sub2(tmp)
  outvar = outvar + tmp + func1()  ! (invar + 1) + 1 + 6
  call sub3(tmp)
  outvar = outvar + tmp + func2()  ! (invar + 8) + 4 + 2
  tmp = func3()
  outvar = outvar + tmp  ! (invar + 14) + 2
end subroutine routine_external_stmt

subroutine routine_call_external_stmt(invar, outvar)
  implicit none
  integer, intent(in) :: invar
  integer, intent(out) :: outvar

  interface
    subroutine other_external_subroutine(outvar)
      integer, intent(out) :: outvar
    end subroutine other_external_subroutine
  end interface

  interface
    function other_external_function()
      integer :: other_external_function
    end function other_external_function
  end interface

  call routine_external_stmt(invar, external_subroutine, external_subroutine, other_external_subroutine, &
                            &outvar, other_external_function, external_function, external_function)

contains

  subroutine external_subroutine(outvar)
    implicit none
    integer, intent(out) :: outvar
    outvar = 1
  end subroutine external_subroutine

  function external_function()
    implicit none
    integer :: external_function
    external_function = 2
  end function external_function

end subroutine routine_call_external_stmt
    """.strip()

    source = Sourcefile.from_source(fcode, frontend=frontend)
    routine = source['routine_external_stmt']
    assert len(routine.arguments) == 8

    for decl in FindNodes(ir.ProcedureDeclaration).visit(routine.spec):
        # Is the EXTERNAL attribute set?
        assert decl.external
        for var in decl.symbols:
            assert isinstance(var, sym.ProcedureSymbol)
            assert isinstance(var.type.dtype, ProcedureType)
            assert var.type.external is True
            assert var.type.dtype.procedure == BasicType.DEFERRED
            if 'sub' in var.name:
                assert not var.type.dtype.is_function
                assert var.type.dtype.return_type is None
            else:
                assert var.type.dtype.is_function
                assert var.type.dtype.return_type.compare(SymbolAttributes(BasicType.INTEGER))

    # Generate code, compile and load
    extpath = tmp_path/(f'subroutine_routine_external_{frontend}.f90')
    extpath.write_text(fcode_external)
    filepath = tmp_path/(f'subroutine_routine_external_stmt_{frontend}.f90')
    source.path = filepath
    lib = jit_compile_lib([source, extpath], path=tmp_path, name='subroutine_external')
    function = lib.routine_call_external_stmt

    outvar = function(7)
    assert outvar == 23
    clean_test(filepath)


@pytest.mark.parametrize('frontend', available_frontends())
def test_subroutine_interface(tmp_path, frontend, header_path):
    """
    Test auto-generation of an interface block for a given subroutine.
    """
    fcode = """
subroutine test_subroutine_interface (in1, in2, in3, out1, out2)
  use header, only: jprb
  IMPLICIT NONE
  integer, intent(in) :: in1, in2
  real(kind=jprb), intent(in) :: in3(in1, in2)
  real(kind=jprb), intent(out) :: out1, out2
  integer :: localvar
  localvar = in1 + in2
  out1 = real(localvar, kind=jprb)
  out2 = out1 + 2.
end subroutine
"""
    if frontend == OMNI:
        # Generate xmod
        Sourcefile.from_file(header_path, frontend=frontend, xmods=[tmp_path])

    routine = Subroutine.from_source(fcode, xmods=[tmp_path], frontend=frontend)

    if frontend == OMNI:
        assert fgen(routine.interface).strip() == """
INTERFACE
  SUBROUTINE test_subroutine_interface (in1, in2, in3, out1, out2)
    USE header, ONLY: jprb
    IMPLICIT NONE
    INTEGER, INTENT(IN) :: in1
    INTEGER, INTENT(IN) :: in2
    REAL(KIND=selected_real_kind(13, 300)), INTENT(IN) :: in3(in1, in2)
    REAL(KIND=selected_real_kind(13, 300)), INTENT(OUT) :: out1
    REAL(KIND=selected_real_kind(13, 300)), INTENT(OUT) :: out2
  END SUBROUTINE test_subroutine_interface
END INTERFACE
""".strip()
    else:
        assert fgen(routine.interface).strip() == """
INTERFACE
  SUBROUTINE test_subroutine_interface (in1, in2, in3, out1, out2)
    USE header, ONLY: jprb
    IMPLICIT NONE
    INTEGER, INTENT(IN) :: in1, in2
    REAL(KIND=jprb), INTENT(IN) :: in3(in1, in2)
    REAL(KIND=jprb), INTENT(OUT) :: out1, out2
  END SUBROUTINE test_subroutine_interface
END INTERFACE
""".strip()


@pytest.mark.parametrize('frontend', available_frontends())
def test_mixed_declaration_interface(frontend):
    """
    A simple test to catch and shame mixed declarations.
    """
    fcode = """
subroutine valid_fortran(i, m)
   integer :: i, j, m
   integer :: k,l
end subroutine valid_fortran
"""

    with pytest.raises(AssertionError) as error:
        routine = Subroutine.from_source(fcode, frontend=frontend)
        assert isinstance(routine.body, ir.Section)
        assert isinstance(routine.spec, ir.Section)
        _ = routine.interface

    assert 'Declarations must have intents' in str(error.value)


@pytest.mark.parametrize('frontend', available_frontends())
def test_routine_prefix(frontend):
    """ Test matching of prefix attributes for subroutines """
    fcode = """
pure elemental subroutine my_routine(x, y)
  implicit none
  integer(kind=8), intent(inout) :: x, y

  x = x + y
end subroutine my_routine
"""
    routine = Subroutine.from_source(fcode, frontend=frontend)

    assert routine.name == 'my_routine'
    assert len(routine.prefix) == 2
    assert routine.prefix == ('PURE', 'ELEMENTAL')

    assert isinstance(routine.body.body[-1], ir.Assignment)
    assert routine.body.body[-1].lhs == 'x'
    assert routine.body.body[-1].rhs == 'x + y'


@pytest.mark.parametrize('frontend', available_frontends(
    skip=[(OMNI, 'OMNI frontend interface does not provide interfaces')]
))
def test_routine_bind(frontend, tmp_path):
    """ Test matching of 'bind" suffix for subroutines in interfaces """
    fcode = """
module my_module
  implicit none

  interface
    subroutine my_routine(x, y) bind(C, name='my_routine_c')
      use, intrinsic :: iso_c_binding
      integer(kind=c_int), intent(inout) :: x, y
    end subroutine my_routine
  end interface

contains

  subroutine my_routine(x, y)
    integer(kind=4), intent(inout) :: x, y

    x = x + y
  end subroutine my_routine
end module my_module
"""
    module = Sourcefile.from_source(fcode, frontend=frontend, xmods=[tmp_path])['my_module']

    routine = module['my_routine']
    intf_routine = module.interface_map['my_routine'].body[0]
    assert isinstance(intf_routine.bind, sym.StringLiteral)
    assert intf_routine.bind == 'my_routine_c'
    assert "BIND(c, name='my_routine_c')" in fgen(intf_routine)

    assert intf_routine == routine

    assert isinstance(routine.body.body[-1], ir.Assignment)
    assert routine.body.body[-1].lhs == 'x'
    assert routine.body.body[-1].rhs == 'x + y'


@pytest.mark.parametrize('frontend', available_frontends())
def test_subroutine_stmt_func(tmp_path, frontend):
    """
    Test the correct identification of statement functions
    """
    fcode = """
subroutine subroutine_stmt_func(a, b)
    implicit none
    integer, intent(in) :: a
    integer, intent(out) :: b
    integer :: array(a)
    integer :: i, j, plus, minus
    plus(i, j) = i + j
    minus(i, j) = i - j
    integer :: mult
    integer :: tmp
    mult(i, j) = i * j

    array(a) = a
    tmp = plus(a, 5)
    tmp = minus(tmp, 1)
    b = mult(2, tmp)
end subroutine subroutine_stmt_func
    """
    routine = Subroutine.from_source(fcode, frontend=frontend)
    routine.name += f'_{frontend!s}'

    for assignment in FindNodes(ir.Assignment).visit(routine.body):
        assert assignment.source is not None

    if frontend != OMNI:
        stmt_func_decls = {decl.variable: decl for decl in FindNodes(ir.StatementFunction).visit(routine.spec)}
        assert len(stmt_func_decls) == 3

        for name in ('plus', 'minus', 'mult'):
            var = routine.variable_map[name]
            assert isinstance(var, sym.ProcedureSymbol)
            assert isinstance(var.type.dtype, ProcedureType)
            assert var.type.dtype.procedure is stmt_func_decls[var]
            assert stmt_func_decls[var].source is not None

    filepath = tmp_path/f'{routine.name}.f90'
    function = jit_compile(routine, filepath=filepath, objname=routine.name)
    assert function(3) == 14
    clean_test(filepath)



# (C) Copyright 2018- ECMWF.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
"""
Utilities to facilitate Just-in-Time compilation for testing purposes.
"""
from pathlib import Path
from multiprocessing import get_context
import os
import traceback

from loki.backend import fgen
from loki.jit_build.builder import Builder
from loki.jit_build.compiler import compile_and_load
from loki.jit_build.lib import Lib
from loki.jit_build.obj import Obj
from loki.ir import Section
from loki.module import Module
from loki.sourcefile import Sourcefile
from loki.subroutine import Subroutine
from loki.tools import as_tuple, gettempdir, filehash


__all__ = ['jit_compile', 'jit_compile_and_run', 'jit_compile_lib', 'run_isolated', 'clean_test']


_f90wrap_kind_map = Path(__file__).parent.parent/'tests/kind_map'


def _run_isolated_worker(conn, target, args, kwargs, exit_after_result):
    """
    Execute a callable in a subprocess and report the outcome through a queue.
    """
    try:
        conn.send(('result', target(*args, **kwargs)))
    except Exception:  # pylint: disable=broad-exception-caught
        conn.send(('exception', traceback.format_exc()))
    finally:
        conn.close()
    if exit_after_result:
        os._exit(0)  # pylint: disable=protected-access


def run_isolated(target, *args, multiprocessing_context='fork', exit_after_result=False,
                 timeout=None, **kwargs):
    """
    Execute ``target`` in a short-lived subprocess and return its result.

    This is useful for tests that JIT-compile and import native extension modules,
    where unloading all linked Fortran/Python wrapper state from the current process
    is not reliable. Python exceptions raised by ``target`` are re-raised as
    :any:`RuntimeError` with the child traceback; native crashes or explicit
    non-zero exits are reported via the child process exit code.

    Parameters
    ----------
    target : callable
        The callable to execute in the child process.
    multiprocessing_context : str, optional
        Multiprocessing start method. Use ``'fork'`` for low-overhead isolation
        when inherited process state is safe, or ``'spawn'`` when the child must
        start without inherited native extension modules.
    exit_after_result : bool, optional
        Exit the child process immediately after reporting the result, bypassing
        interpreter shutdown. This is useful for native extension tests where
        finalizers can crash after successful execution.
    timeout : int or float, optional
        Maximum time in seconds to wait for the isolated child process.
    """
    ctx = get_context(multiprocessing_context)
    parent_conn, child_conn = ctx.Pipe(duplex=False)
    process = ctx.Process(
        target=_run_isolated_worker, args=(child_conn, target, args, kwargs, exit_after_result)
    )
    process.start()
    child_conn.close()
    process.join(timeout=timeout)

    if process.is_alive():
        process.terminate()
        process.join()
        parent_conn.close()
        raise RuntimeError(f'Isolated process timed out after {timeout} seconds')

    try:
        has_payload = parent_conn.poll()
        kind, payload = parent_conn.recv() if has_payload else (None, None)
    except EOFError:
        kind = payload = None
    parent_conn.close()

    if kind == 'exception':
        raise RuntimeError(payload)
    if process.exitcode != 0:
        raise RuntimeError(f'Isolated process failed with exit code {process.exitcode}')
    if kind == 'result':
        return payload
    raise RuntimeError('Unexpected return form child process - something has gone terribly wrong!')


def _jit_compile_and_run(source, args, kwargs, filepath, objname):
    """
    JIT-compile ``source`` and execute the selected entry point.
    """
    function = jit_compile(source, filepath=filepath, objname=objname)
    result = function(*args, **kwargs)
    return result, args, kwargs


def _copy_back_arguments(original, updated):
    """
    Copy updated array-like argument values back into the caller objects.
    """
    for original_value, updated_value in zip(original, updated):
        if hasattr(original_value, '__setitem__') and hasattr(original_value, 'shape'):
            original_value[...] = updated_value


def jit_compile_and_run(source, *args, filepath=None, objname=None, isolated=True,
                        multiprocessing_context='fork', exit_after_result=False,
                        timeout=10, **kwargs):
    """
    JIT-compile a source item and execute the selected entry point.

    The complete compile, load, and execution cycle can optionally run in a
    short-lived subprocess to isolate native extension module state from the
    parent test process. See :any:`run_isolated` for available options for arguments.

    Parameters
    ----------
    source : :any:`Sourcefile` or :any:`Module` or :any:`Subroutine`
        The item to compile and load.
    *args : tuple
        Positional arguments passed to the compiled entry point.
    filepath : str or :any:`Path`, optional
        Path of the source file to write.
    objname : str, optional
        Name of the compiled object to execute. Defaults to ``source.name``
        when available.
    isolated : bool, optional
        Run the compile-and-execute cycle via :any:`run_isolated` when enabled
        (default: ``True``).
    multiprocessing_context : str, optional
        Multiprocessing start method used for isolated execution.
    exit_after_result : bool, optional
        Exit the isolated child immediately after reporting the result.
    timeout : int or float, optional
        Maximum time in seconds to wait for isolated execution (default: 10).
    **kwargs : dict
        Keyword arguments passed to the compiled entry point.
    """
    objname = objname or getattr(source, 'name', None)
    if isolated:
        result, updated_args, updated_kwargs = run_isolated(
            _jit_compile_and_run, source, args, kwargs, filepath, objname,
            multiprocessing_context=multiprocessing_context, exit_after_result=exit_after_result,
            timeout=timeout
        )
        _copy_back_arguments(args, updated_args)
        _copy_back_arguments(kwargs.values(), updated_kwargs.values())
        return result
    result, _, _ = _jit_compile_and_run(source, args, kwargs, filepath, objname)
    return result


def jit_compile(source, filepath=None, objname=None):
    """
    Generate, Just-in-Time compile and load a given item
    for interactive execution.

    Parameters
    ----------
    source : :any:`Sourcefile` or :any:`Module` or :any:`Subroutine`
        The item to compile and load
    filepath : str or :any:`Path`, optional
        Path of the source file to write (default: hashed name in :any:`gettempdir()`)
    objname : str, optional
        Return a specific object (module or subroutine) in :attr:`source`
    """
    if isinstance(source, Sourcefile):
        filepath = source.path if filepath is None else Path(filepath)
        if filepath is None:
            filepath = Path(gettempdir()/filehash(source, prefix='', suffix='.f90'))
        source.write(path=filepath)
    else:
        source = fgen(source)
        if filepath is None:
            filepath = gettempdir()/filehash(source, prefix='', suffix='.f90')
        else:
            filepath = Path(filepath)
        Sourcefile(filepath).write(source=source)

    pymod = compile_and_load(filepath, cwd=str(filepath.parent), f90wrap_kind_map=_f90wrap_kind_map)

    if objname:
        return getattr(pymod, objname)
    return pymod


def jit_compile_lib(sources, path, name, wrap=None, builder=None):
    """
    Generate, just-in-time compile and load a set of items into a
    library and import dynamically into the Python runtime.

    Parameters
    ----------
    source : list
        Source items or filepaths to compile and add to lib
    path : str or :any:`Path`
        Basepath for on-the-fly creation of source files
    name : str
        Name of created lib
    wrap : list, optional
        File names to pass to ``f90wrap``. Defaults to list of source files.
    builder : :any:`Builder`, optional
        Builder object to use for lib compilation and linking
    """
    if builder is None:
        builder_provided = False
        builder = Builder(source_dirs=path, build_dir=path)
    else:
        builder_provided = True
    sourcefiles = []

    for source in sources:
        if isinstance(source, (str, Path)):
            sourcefiles.append(source)

        if isinstance(source, Sourcefile):
            filepath = source.path or path/f'{source.name}.f90'
            source.write(path=filepath)
            sourcefiles.append(source.path)

        elif isinstance(source, (Module, Subroutine)):
            filepath = path/f'{source.name}.f90'
            source = Sourcefile(filepath, ir=Section(body=as_tuple(source)))
            source.write(path=filepath)
            sourcefiles.append(source.path)

    objects = [Obj(source_path=s) for s in sourcefiles]
    lib = Lib(name=name, objs=objects, shared=False)
    lib.build(builder=builder)
    wrap = wrap or sourcefiles
    pymod = lib.wrap(modname=name, sources=wrap, builder=builder, kind_map=_f90wrap_kind_map)
    if not builder_provided:
        Obj.clear_cache()
    return pymod


def clean_test(filepath):
    """
    Clean test directory based on JIT'ed source file.
    """
    file_list = [
        filepath.with_suffix('.f90'), filepath.with_suffix('.o'),
        filepath.with_suffix('.py'), filepath.parent/'f90wrap_toplevel.f90',
        filepath.with_suffix('.mod'), filepath.with_suffix('.xmod')
    ]
    for f in file_list:
        if f.exists():
            f.unlink()
    for sofile in filepath.parent.glob(f'_{filepath.stem}.*.so'):
        sofile.unlink()
    f90wrap_path = filepath.parent/f'f90wrap_{filepath.name}'
    if f90wrap_path.exists():
        f90wrap_path.unlink()



# (C) Copyright 2018- ECMWF.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from loki.batch import Transformation
from loki.transformations.remove_code import do_remove_dead_code

from loki.ir import pragmas_attached, CallStatement, FindNodes, is_loki_pragma
from loki.tools.util import as_tuple, CaseInsensitiveDict, OrderedSet
from loki.transformations.inline.constants import inline_constant_parameters
from loki.transformations.inline.functions import (
    inline_elemental_functions, inline_statement_functions
)
from loki.transformations.inline.procedures import (
    inline_internal_procedures, inline_marked_subroutines,
    resolve_sequence_association_for_inlined_calls
)


__all__ = ['InlineTransformation']


class InlineTransformation(Transformation):
    """
    :any:`Transformation` class to apply several types of source inlining
    when batch-processing large source trees via the :any:`Scheduler`.

    Parameters
    ----------
    inline_constants : bool
        Replace instances of variables with known constant values by
        :any:`Literal` (see :any:`inline_constant_parameters`); default: False.
    inline_elementals : bool
        Replaces :any:`InlineCall` expression to elemental functions
        with the called function's body (see :any:`inline_elemental_functions`);
        default: True.
    inline_stmt_funcs: bool
        Replaces  :any:`InlineCall` expression to statement functions
        with the corresponding rhs of the statement function if
        the statement function declaration is available; default: False.
    inline_internals : bool
        Inline internal procedure (see :any:`inline_internal_procedures`);
        default: False.
    inline_marked : bool
        Inline :any:`Subroutine` objects marked by pragma annotations
        (see :any:`inline_marked_subroutines`); default: True.
    remove_dead_code : bool
        Perform dead code elimination, where unreachable branches are
        trimmed from the code (see :any:`dead_code_elimination`); default: True
    allowed_aliases : tuple or list of str or :any:`Expression`, optional
        List of variables that will not be renamed in the parent scope during
        internal and pragma-driven inlining.
    adjust_imports : bool
        Adjust imports by removing the symbol of the inlined routine or adding
        imports needed by the imported routine (optional, default: True)
    external_only : bool, optional
        Do not replace variables declared in the local scope when
        inlining constants (default: True)
    resolve_sequence_association: bool
        Resolve sequence association for routines that contain calls to inline (default: False)
    """

    # Ensure correct recursive inlining by traversing from the leaves
    reverse_traversal = True

    # This transformation will potentially change the edges in the callgraph
    creates_items = False

    def __init__(
            self, inline_constants=False, inline_elementals=True,
            inline_stmt_funcs=False, inline_internals=False,
            inline_marked=True, remove_dead_code=True,
            allowed_aliases=None, adjust_imports=True,
            external_only=True, resolve_sequence_association=False
    ):
        self.inline_constants = inline_constants
        self.inline_elementals = inline_elementals
        self.inline_stmt_funcs = inline_stmt_funcs
        self.inline_internals = inline_internals
        self.inline_marked = inline_marked
        self.remove_dead_code = remove_dead_code
        self.allowed_aliases = allowed_aliases
        self.adjust_imports = adjust_imports
        self.external_only = external_only
        self.resolve_sequence_association = resolve_sequence_association
        if self.inline_marked:
            self.creates_items = True

    def transform_subroutine(self, routine, **kwargs):

        # Resolve sequence association in calls that are about to be inlined.
        # This step runs only if all of the following hold:
        # 1) it is requested by the user
        # 2) inlining of "internals" or "marked" routines is activated
        # 3) there is an "internal" or "marked" procedure to inline.
        if self.resolve_sequence_association:
            resolve_sequence_association_for_inlined_calls(
                routine, self.inline_internals, self.inline_marked
            )

        # Replace constant parameter variables with explicit values
        if self.inline_constants:
            inline_constant_parameters(routine, external_only=self.external_only)

        # Inline elemental functions
        if self.inline_elementals:
            inline_elemental_functions(routine)

        # Inline Statement Functions
        if self.inline_stmt_funcs:
            inline_statement_functions(routine)

        # Inline internal (contained) procedures
        if self.inline_internals:
            inline_internal_procedures(routine, allowed_aliases=self.allowed_aliases)

        # Inline explicitly pragma-marked subroutines
        if self.inline_marked:
            inline_marked_subroutines(
                routine, allowed_aliases=self.allowed_aliases,
                adjust_imports=self.adjust_imports
            )

        # After inlining, attempt to trim unreachable code paths
        if self.remove_dead_code:
            do_remove_dead_code(routine)

    def plan_subroutine(self, routine, **kwargs):

        if not self.inline_marked:
            return

        item = kwargs.get('item')
        sub_sgraph = kwargs.get('sub_sgraph', None)
        successors = sub_sgraph.successors(item) if sub_sgraph is not None else ()
        successor_map = CaseInsensitiveDict(
            (successor.local_name, successor) for successor in successors
        )
        # look for call statements with pragmas attached
        with pragmas_attached(routine, node_type=CallStatement):
            inline_calls = OrderedSet()
            not_inline_calls = OrderedSet()
            calls = FindNodes(CallStatement).visit(routine.ir)
            # for all calls sort those having '!$loki inline' and those not having it
            for call in calls:
                if is_loki_pragma(call.pragma, starts_with='inline'):
                    inline_calls.add(str(call.name).lower())
                else:
                    not_inline_calls.add(str(call.name).lower())
        # Determine the list of routines that will be completely inlined and therefore no longer be dependencies
        # of the current item. If calls to the same routine remain non-inlined, the dependency remains, too.
        removed_calls = inline_calls - not_inline_calls
        rename_map = CaseInsensitiveDict(
                (s.name, s.type.use_name if s.type.use_name else s.name)
                for imprt in reversed(getattr(routine, 'imports', ()))
                for s in imprt.symbols or [r[1] for r in imprt.rename_list or ()]
                )
        inline_items = [successor_map[rename_map.get(call, call)] for call in removed_calls
                # this shouldn't be necessary, however, if for example a call is marked as to be inlined
                # within a loki remove pragma region this could otherwise end up throwing an error
                if rename_map.get(call, call) in successor_map]
        # Add fully inlined dependencies to the 'removed_dependencies' list in the plan data to indicate that
        # they will no longer be dependents. At the same time add any dependencies of inlined successors
        # to the current item's dependencies as these will be inherited as dependents (unless they are also
        # inlined)
        if inline_items:
            item.plan_data.setdefault('removed_dependencies', ())
            item.plan_data.setdefault('additional_dependencies', ())
            item.plan_data['removed_dependencies'] += as_tuple(inline_items)
            additional_dep = ()
            for inline_item in inline_items:
                inlined_successors = sub_sgraph.successors(inline_item) + \
                        inline_item.plan_data.get('additional_dependencies', ())
                for inlined_successor in inlined_successors:
                    if inlined_successor not in inline_item.plan_data.get('removed_dependencies', ()):
                        additional_dep += (inlined_successor,)
            item.plan_data['additional_dependencies'] += as_tuple(OrderedSet(additional_dep))



# (C) Copyright 2018- ECMWF.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

"""
Specific test battery for the OMNI parser frontend.
"""

import pytest

from loki import Module, Subroutine
from loki.backend import fgen
from loki.expression import symbols as sym
from loki.frontend import OMNI, HAVE_OMNI
from loki.ir import nodes as ir, FindNodes, FindVariables


@pytest.mark.skipif(not HAVE_OMNI, reason='Test tequires OMNI frontend.')
def test_derived_type_definitions(tmp_path):
    """ Test correct parsing of derived type declarations. """
    fcode = """
module omni_derived_type_mod
  type explicit
      real(kind=8) :: scalar, vector(3), matrix(3, 3)
  end type explicit

  type deferred
      real(kind=8), allocatable :: scalar, vector(:), matrix(:, :)
  end type deferred

  type ranged
      real(kind=8) :: scalar, vector(1:3), matrix(0:3, 0:3)
  end type ranged
end module omni_derived_type_mod
"""
    # Parse the source and validate the IR
    module = Module.from_source(fcode, frontend=OMNI, xmods=[tmp_path])

    assert len(module.typedefs) == 3
    explicit_symbols = FindVariables(unique=False).visit(module['explicit'].body)
    assert explicit_symbols == ('scalar', 'vector(3)', 'matrix(3, 3)')

    deferred_symbols = FindVariables(unique=False).visit(module['deferred'].body)
    assert deferred_symbols == ('scalar', 'vector(:)', 'matrix(:, :)')

    ranged_symbols = FindVariables(unique=False).visit(module['ranged'].body)
    assert ranged_symbols == ('scalar', 'vector(3)', 'matrix(0:3, 0:3)')


@pytest.mark.skipif(not HAVE_OMNI, reason='Test tequires OMNI frontend.')
def test_array_dimensions(tmp_path):
    """ Test correct parsing of derived type declarations. """
    fcode = """
subroutine omni_array_indexing(n, a, b)
  integer, intent(in) :: n
  real(kind=8), intent(inout) :: a(3), b(n)
  real(kind=8) :: c(n, n)
  real(kind=8) :: d(1:n, 0:n)

  a(:) = 11.
  b(1:n) = 42.
  c(2:n, 0:n) = 66.
  d(:, 0:n) = 68.
end subroutine omni_array_indexing
"""
    # Parse the source and validate the IR
    routine = Subroutine.from_source(fcode, frontend=OMNI, xmods=[tmp_path])

    # OMNI separate declarations per variable
    decls = FindNodes(ir.VariableDeclaration).visit(routine.spec)
    assert len(decls) == 5
    assert decls[0].symbols == ('n',)
    assert decls[1].symbols == ('a(3)',)
    assert decls[2].symbols == ('b(n)',)
    assert decls[3].symbols == ('c(n, n)',)
    assert decls[4].symbols == ('d(n, 0:n)',)

    assigns = FindNodes(ir.Assignment).visit(routine.body)
    assert len(assigns) == 4
    assert assigns[0].lhs == 'a(:)'
    assert assigns[1].lhs == 'b(1:n)'
    assert assigns[2].lhs == 'c(2:n, 0:n)'
    assert assigns[3].lhs == 'd(:, 0:n)'


@pytest.mark.skipif(not HAVE_OMNI, reason='Test tequires OMNI frontend.')
def test_labelled_format_with_quoted_items():
    """Test parsing of labelled FORMAT statements with quoted comma literals."""
    fcode = """
subroutine omni_labelled_format()
  implicit none
1003 format(5x,'NUMPROC=',i0,', NUMOMP=',i0,', NGPTOTG=',i0,', NPROMA=',i0,', NGPBLKS=',i0)
end subroutine omni_labelled_format
"""
    routine = Subroutine.from_source(fcode, frontend=OMNI)
    stmt = FindNodes(ir.FormatStmt).visit(routine.body)[0]

    assert stmt.label == '1003'
    assert stmt.values == (
        '5x', sym.StringLiteral('NUMPROC='), 'i0', sym.StringLiteral(', NUMOMP='), 'i0',
        sym.StringLiteral(', NGPTOTG='), 'i0', sym.StringLiteral(', NPROMA='), 'i0',
        sym.StringLiteral(', NGPBLKS='), 'i0'
    )
    assert fgen(routine.body) == (
        "1003 FORMAT(5x, 'NUMPROC=', i0, ', NUMOMP=', i0, ', NGPTOTG=', i0, "
        "', NPROMA=', i0, ', NGPBLKS=', i0)"
    )



# (C) Copyright 2018- ECMWF.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
"""
Utilities for Loki scheduler tests.

The overall structure of the source projects for testing is:

 - projA:
   - include
     - another_l1.intfb.h
     - another_l2.intfb.h
   - source
     - another_l1
     - another_l2
   - module
     - header_mod
       * header_type
     - driverA_mod
     - kernelA_mod
     - compute_l1_mod
     - compute_l2_mod
     - driverB_mod
     - kernelB_mod
     - driverC_mod
     - kernelC_mod
     - driverD_mod
     - kernelD_mod

 - projB:
   - external
     - ext_driver_mod
   - module
     - ext_kernel_mod

 - projC:
   - util
     - proj_c_util_mod
       * routine_one
       * routine_two
"""
from pathlib import Path
import re

import pytest

from loki.frontend import FP


@pytest.fixture(scope='module', name='here')
def fixture_here():
    return Path(__file__).parent


@pytest.fixture(scope='module', name='testdir')
def fixture_testdir(here):
    return here.parent.parent/'tests'


@pytest.fixture(name='config')
def fixture_config():
    """
    Default configuration dict with basic options.
    """
    return {
        'default': {
            'mode': 'idem',
            'role': 'kernel',
            'expand': True,
            'strict': True,
            'disable': ['abort'],
            'enable_imports': True,
        },
        'routines': {}
    }


@pytest.fixture(name='frontend')
def fixture_frontend():
    """
    Frontend to use.

    Not parametrizing the tests as the scheduler functionality should be
    independent from the specific frontend used. Cannot use OMNI for this
    as not all tests have dependencies fully resolved.
    """
    return FP


class VisGraphWrapper:
    """
    Testing utility to parse the generated callgraph visualisation.
    """

    _re_nodes = re.compile(r'\s*\"?(?P<node>[\w%#./-]+)\"? \[colo', re.IGNORECASE)
    _re_edges = re.compile(r'\s*\"?(?P<parent>[\w%#./-]+)\"? -> \"?(?P<child>[\w%#./-]+)\"?', re.IGNORECASE)

    def __init__(self, path):
        with Path(path).open('r') as f:
            self.text = f.read()

    @property
    def nodes(self):
        return list(self._re_nodes.findall(self.text))

    @property
    def edges(self):
        return list(self._re_edges.findall(self.text))



# (C) Copyright 2018- ECMWF.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

# pylint: disable=too-many-lines

from xml.etree.ElementTree import ParseError

import pytest

from loki import (
    Array, BasicType, CaseInsensitiveDict, DerivedType, ProcedureType,
    Scalar, Sourcefile, Subroutine, available_frontends, fexprgen,
    graphviz_present
)
from loki.batch import (
    ExternalItem, InterfaceItem, ModuleItem, ProcedureItem, ProcessingStrategy,
    Scheduler, SchedulerConfig, Transformation, TypeDefItem
)
from loki.expression import ProcedureSymbol
from loki.frontend import FP, HAVE_FP, OMNI, REGEX
from loki.ir import FindInlineCalls, FindNodes, FindVariables, nodes as ir

from .conftest import VisGraphWrapper

pytestmark = pytest.mark.skipif(not HAVE_FP, reason='Fparser not available')


@pytest.fixture(name='proj_typebound_dependencies')
def fixture_proj_typebound_dependencies():
    return {
        '#driver': (
            'typebound_item',
            'typebound_item#some_type%other_routine',
            'typebound_item#some_type%some_routine',
            'typebound_header',
            'typebound_header#header_type%member_routine',
            'typebound_header#header_type%routine',
            'typebound_other#other_type',
            'typebound_other#other_type%member',
            'typebound_other#other_type%var%member_routine',
        ),
        'typebound_item': ('typebound_header',),
        'typebound_item#some_type': (),
        'typebound_item#some_type%some_routine': ('typebound_item#some_routine',),
        'typebound_item#some_type%other_routine': ('typebound_item#other_routine',),
        'typebound_item#some_type%routine': ('typebound_item#module_routine',),
        'typebound_item#some_type%routine1': ('typebound_item#routine1',),
        'typebound_item#some_type%routine2': ('typebound_item#routine',),
        'typebound_item#routine': (
            'typebound_item#some_type',
            'typebound_item#some_type%some_routine',
        ),
        'typebound_item#routine1': (
            'typebound_item#module_routine',
            'typebound_item#some_type'
        ),
        'typebound_item#some_routine': (
            'typebound_item#some_type',
            'typebound_item#some_type%routine',
        ),
        'typebound_item#other_routine': (
            'typebound_item#some_type',
            'typebound_item#some_type%routine1',
            'typebound_item#some_type%routine2',
            'typebound_header#abor1'
        ),
        'typebound_item#module_routine': (),
        'typebound_header': (),
        'typebound_header#header_type': (),
        'typebound_header#header_type%member_routine': ('typebound_header#header_member_routine',),
        'typebound_header#header_member_routine': ('typebound_header#header_type',),
        'typebound_header#header_type%routine': (
            'typebound_header#header_type%routine_real',
            'typebound_header#header_type%routine_integer'
        ),
        'typebound_header#header_type%routine_real': ('typebound_header#header_routine_real',),
        'typebound_header#header_routine_real': ('typebound_header#header_type',),
        'typebound_header#header_type%routine_integer': ('typebound_header#routine_integer',),
        'typebound_header#routine_integer': ('typebound_header#header_type',),
        'typebound_header#abor1': (),
        'typebound_other#other_type': ('typebound_header#header_type',),
        'typebound_other#other_type%member': ('typebound_other#other_member',),
        'typebound_other#other_member': (
            'typebound_header#header_member_routine',
            'typebound_other#other_type',
            'typebound_other#other_type%var%member_routine'
        ),
        'typebound_other#other_type%var%member_routine': ('typebound_header#header_type%member_routine',)
    }


@pytest.fixture(name='loki_69_dir')
def fixture_loki_69_dir(testdir):
    """
    Fixture to write test file for LOKI-69 test.
    """
    fcode = """
subroutine random_call_0(v_out,v_in,v_inout)
implicit none

    real(kind=jprb),intent(in)  :: v_in
    real(kind=jprb),intent(out)  :: v_out
    real(kind=jprb),intent(inout)  :: v_inout


end subroutine random_call_0

!subroutine random_call_1(v_out,v_in,v_inout)
!implicit none
!
!  real(kind=jprb),intent(in)  :: v_in
!  real(kind=jprb),intent(out)  :: v_out
!  real(kind=jprb),intent(inout)  :: v_inout
!
!
!end subroutine random_call_1

subroutine random_call_2(v_out,v_in,v_inout)
implicit none

    real(kind=jprb),intent(in)  :: v_in
    real(kind=jprb),intent(out)  :: v_out
    real(kind=jprb),intent(inout)  :: v_inout


end subroutine random_call_2

subroutine test(v_out,v_in,v_inout,some_logical)
implicit none

    real(kind=jprb),intent(in   )  :: v_in
    real(kind=jprb),intent(out  )  :: v_out
    real(kind=jprb),intent(inout)  :: v_inout

    logical,intent(in)             :: some_logical

    v_inout = 0._jprb
    if(some_logical)then
        call random_call_0(v_out,v_in,v_inout)
    endif

    if(some_logical) call random_call_2

end subroutine test
    """.strip()

    dirname = testdir/'loki69'
    dirname.mkdir(exist_ok=True)
    filename = dirname/'test.F90'
    filename.write_text(fcode)
    yield dirname
    try:
        filename.unlink()
        dirname.rmdir()
    except FileNotFoundError:
        pass


def test_scheduler_enrichment(testdir, config, frontend, tmp_path):
    projA = testdir/'sources/projA'

    scheduler = Scheduler(
        paths=projA, includes=projA/'include', config=config,
        seed_routines=['driverA'], frontend=frontend, xmods=[tmp_path]
    )

    for item in scheduler.sgraph:
        if not isinstance(item, ProcedureItem):
            continue
        dependency_map = CaseInsensitiveDict(
            (item_.local_name, item_) for item_ in scheduler.sgraph.successors(item)
        )
        for call in FindNodes(ir.CallStatement).visit(item.ir.body):
            if call_item := dependency_map.get(str(call.name)):
                assert call.routine is call_item.ir


@pytest.mark.parametrize('seed', ['driverA', 'driverA_mod#driverA'])
def test_scheduler_definitions(testdir, config, frontend, seed, tmp_path):
    """
    Create a simple task graph and inject type info via `definitions`.
    """
    projA = testdir/'sources/projA'

    header = Sourcefile.from_file(projA/'module/header_mod.f90', frontend=frontend)

    scheduler = Scheduler(
        paths=projA, definitions=header['header_mod'], includes=projA/'include',
        config=config, seed_routines=[seed], frontend=frontend, xmods=[tmp_path]
    )

    driver = scheduler.item_factory.item_cache['drivera_mod#drivera'].ir
    call = FindNodes(ir.CallStatement).visit(driver.body)[0]
    assert call.arguments[0].parent.type.dtype.typedef is not BasicType.DEFERRED
    assert fexprgen(call.arguments[0].shape) == '(:,)'
    assert call.arguments[1].parent.type.dtype.typedef is not BasicType.DEFERRED
    assert fexprgen(call.arguments[1].shape) == '(3, 3)'


def test_scheduler_module_dependency(testdir, config, frontend, tmp_path):
    """
    Ensure dependency chasing is done correctly, even with surboutines
    that do not match module names.

    projA: driverC -> kernelC -> compute_l1<replicated> -> compute_l2
                           |
    projC:                 | --> routine_one -> routine_two
    """
    projA = testdir/'sources/projA'
    projC = testdir/'sources/projC'

    scheduler = Scheduler(
        paths=[projA, projC], includes=projA/'include', config=config,
        seed_routines=['driverC_mod#driverC'], frontend=frontend, xmods=[tmp_path]
    )

    expected_dependencies = {
        'driverc_mod#driverc': ('header_mod', 'header_mod#header_type', 'kernelc_mod#kernelc',),
        'kernelc_mod#kernelc': ('compute_l1_mod#compute_l1', 'proj_c_util_mod#routine_one',),
        'compute_l1_mod#compute_l1': ('compute_l2_mod#compute_l2',),
        'compute_l2_mod#compute_l2': (),
        'proj_c_util_mod#routine_one': ('proj_c_util_mod#routine_two',),
        'proj_c_util_mod#routine_two': (),
        'header_mod#header_type': (),
        'header_mod': (),
    }
    assert set(scheduler.items) == set(expected_dependencies)
    assert set(scheduler.dependencies) == {
        (a, b) for a, deps in expected_dependencies.items() for b in deps
    }

    # Ensure that we got the right routines from the module
    assert scheduler['proj_c_util_mod#routine_one'].ir.name == 'routine_one'
    assert scheduler['proj_c_util_mod#routine_two'].ir.name == 'routine_two'


def test_scheduler_module_dependencies_unqualified(testdir, config, frontend, tmp_path):
    """
    Ensure dependency chasing is done correctly for unqualified module imports.

    projA: driverD -> kernelD -> compute_l1<replicated> -> compute_l2
                           |
                    < proj_c_util_mod>
                           |
    projC:                 | --> routine_one -> routine_two
    """
    projA = testdir/'sources/projA'
    projC = testdir/'sources/projC'

    scheduler = Scheduler(
        paths=[projA, projC], includes=projA/'include', config=config,
        seed_routines=['driverD_mod#driverD'], frontend=frontend, xmods=[tmp_path]
    )

    expected_dependencies = {
        'driverd_mod#driverd': ('kerneld_mod#kerneld', 'header_mod', 'header_mod#header_type'),
        'kerneld_mod#kerneld': ('compute_l1_mod#compute_l1', 'proj_c_util_mod#routine_one'),
        'compute_l1_mod#compute_l1': ('compute_l2_mod#compute_l2',),
        'compute_l2_mod#compute_l2': (),
        'proj_c_util_mod#routine_one': ('proj_c_util_mod#routine_two',),
        'proj_c_util_mod#routine_two': (),
        'header_mod#header_type': (),
        'header_mod': (),
    }
    assert set(scheduler.items) == set(expected_dependencies)
    assert set(scheduler.dependencies) == {
        (a, b) for a, deps in expected_dependencies.items() for b in deps
    }

    # Ensure that we got the right routines from the module
    assert scheduler['proj_c_util_mod#routine_one'].ir.name == 'routine_one'
    assert scheduler['proj_c_util_mod#routine_two'].ir.name == 'routine_two'


@pytest.mark.parametrize('strict', [True, False])
def test_scheduler_missing_files(testdir, config, frontend, strict, tmp_path):
    """
    Ensure that ``strict=True`` triggers failure if source paths are
    missing and that ``strict=False`` goes through gracefully.

    projA: driverC -> kernelC -> compute_l1<replicated> -> compute_l2
                           |
    projC:                 < cannot find path >
    """
    projA = testdir/'sources/projA'

    config['default']['strict'] = strict
    scheduler = Scheduler(
        paths=[projA], includes=projA/'include', config=config,
        seed_routines=['driverC_mod#driverC'], frontend=frontend, xmods=[tmp_path]
    )

    expected_dependencies = {
        'driverc_mod#driverc': ('kernelc_mod#kernelc', 'header_mod#header_type', 'header_mod'),
        'kernelc_mod#kernelc': ('compute_l1_mod#compute_l1', 'proj_c_util_mod#routine_one'),
        'compute_l1_mod#compute_l1': ('compute_l2_mod#compute_l2',),
        'compute_l2_mod#compute_l2': (),
        'header_mod#header_type': (),
        'header_mod': (),
        'proj_c_util_mod#routine_one': (),
    }
    assert set(scheduler.items) == set(expected_dependencies)
    assert set(scheduler.dependencies) == {
        (a, b) for a, deps in expected_dependencies.items() for b in deps
    }

    assert isinstance(scheduler['proj_c_util_mod#routine_one'], ExternalItem)
    # Ensure that the missing items are not in the graph
    assert 'proj_c_util_mod#routine_two' not in scheduler.items

    class CheckApply(Transformation):

        def apply(self, source, post_apply_rescope_symbols=False, plan_mode=False, **kwargs):
            assert 'item' in kwargs
            assert not isinstance(kwargs['item'], ExternalItem)
            super().apply(
                source, post_apply_rescope_symbols=post_apply_rescope_symbols,
                plan_mode=plan_mode, **kwargs
            )

    # Check processing with missing items
    if strict:
        with pytest.raises(RuntimeError):
            scheduler.process(CheckApply())
    else:
        scheduler.process(CheckApply())


def test_scheduler_item_dependencies(testdir, tmp_path):
    """
    Make sure children are correct and unique for items
    """
    config = SchedulerConfig.from_dict({
        'default': {'role': 'kernel', 'expand': True, 'strict': True},
        'routines': {
            'driver': {'role': 'driver'},
            'another_driver': {'role': 'driver'}
        }
    })

    proj_multi_mode = testdir/'sources/projHoist'

    scheduler = Scheduler(paths=proj_multi_mode, config=config, xmods=[tmp_path])

    assert tuple(
        call.name for call in scheduler['transformation_module_hoist#driver'].dependencies
    ) == ('kernel1', 'kernel2')
    assert tuple(
        call.name for call in scheduler['transformation_module_hoist#another_driver'].dependencies
    ) == ('kernel1',)
    assert not scheduler['subroutines_mod#kernel1'].dependencies
    assert tuple(
        call.name for call in scheduler['subroutines_mod#kernel2'].dependencies
    ) == ('device1', 'device2')
    assert tuple(
        call.name for call in scheduler['subroutines_mod#device1'].dependencies
    ) == ('device2',)
    assert not scheduler['subroutines_mod#device2'].dependencies


def test_scheduler_loki_69(loki_69_dir, tmp_path):
    """
    Test compliance of scheduler with edge cases reported in LOKI-69.
    """
    config = {
        'default': {
            'expand': True,
            'strict': True,
        },
    }

    scheduler = Scheduler(paths=loki_69_dir, seed_routines=['test'], config=config, xmods=[tmp_path])

    assert sorted(scheduler.item_factory.item_cache.keys()) == [
        '#random_call_0', '#random_call_2', '#test',
        str(loki_69_dir/'test.f90').lower()
    ]
    assert '#random_call_1' not in scheduler.item_factory

    children_map = {
        '#test': ('#random_call_0', '#random_call_2'),
        '#random_call_0': (),
        '#random_call_2': ()
    }
    assert len(scheduler.items) == len(children_map)
    for item in scheduler.items:
        assert set(scheduler.sgraph.successors(item)) == set(children_map[item.name])


@pytest.mark.skipif(not graphviz_present(), reason='Graphviz is not installed')
def test_scheduler_scopes(tmp_path, testdir, config, frontend):
    """
    Test discovery with import renames and duplicate names in separate scopes

    driver ----> kernel1_mod#kernel ----> kernel1_impl#kernel_impl
      |
      +--------> kernel2_mod#kernel ----> kernel2_impl#kernel_impl
    """
    proj = testdir/'sources/projScopes'

    scheduler = Scheduler(paths=proj, seed_routines=['driver'], config=config, frontend=frontend, xmods=[tmp_path])

    expected_dependencies = {
        '#driver': ('kernel1_mod#kernel', 'kernel2_mod#kernel'),
        'kernel1_mod#kernel': ('kernel1_impl', 'kernel1_impl#kernel_impl'),
        'kernel1_impl': (),
        'kernel1_impl#kernel_impl': (),
        'kernel2_mod#kernel': ('kernel2_impl', 'kernel2_impl#kernel_impl'),
        'kernel2_impl': (),
        'kernel2_impl#kernel_impl': (),
    }

    assert set(scheduler.items) == set(expected_dependencies)
    assert set(scheduler.dependencies) == {
        (a, b) for a, deps in expected_dependencies.items()
        for b in deps
    }

    cg_path = tmp_path/'callgraph_scopes'
    scheduler.callgraph(cg_path)

    # Testing of callgraph visualisation
    vgraph = VisGraphWrapper(cg_path)
    assert set(vgraph.nodes) == {n.upper() for n in expected_dependencies}
    assert set(vgraph.edges) == {
        (a.upper(), b.upper()) for a, deps in expected_dependencies.items()
        for b in deps
    }

    cg_path.unlink()
    cg_path.with_suffix('.pdf').unlink()


@pytest.mark.skipif(not graphviz_present(), reason='Graphviz is not installed')
def test_scheduler_typebound(tmp_path, testdir, config, frontend, proj_typebound_dependencies):
    """
    Test correct dependency chasing for typebound procedure calls.

    projTypeBound: driver -> some_type%other_routine -> other_routine -> some_type%routine1 -> routine1
                 | | | | | |                                          |                                |
                 | | | | | |       +- routine <- some_type%routine2 <-+                                +---------+
                 | | | | | |       |                                                                             |
                 | | | | | +--> some_type%some_routine -> some_routine -> some_type%routine -> module_routine  <-+
                 | | | +------> header_type%member_routine -> header_member_routine
                 | | +--------> header_type%routine -> header_type%routine_real -> header_routine_real
                 | |                           |
                 | |                           +---> header_type%routine_integer -> routine_integer
                 | +---------->other_type%member -> other_member -> header_member_routine   <--+
                 |                                                                             |
                 +------------>other_type%var%%member_routine -> header_type%member_routine  --+
    """
    proj = testdir/'sources/projTypeBound'

    scheduler = Scheduler(
        paths=proj, seed_routines=['driver'], config=config,
        full_parse=False, frontend=frontend, xmods=[tmp_path]
    )

    assert set(scheduler.items) == set(proj_typebound_dependencies)
    assert set(scheduler.dependencies) == {
        (a, b) for a, deps in proj_typebound_dependencies.items() for b in deps
    }

    cg_path = tmp_path/'callgraph_typebound'
    scheduler.callgraph(cg_path)

    # Testing of callgraph visualisation
    vgraph = VisGraphWrapper(cg_path)
    assert set(vgraph.nodes) == {n.upper() for n in proj_typebound_dependencies}
    assert set(vgraph.edges) == {
        (a.upper(), b.upper()) for a, deps in proj_typebound_dependencies.items() for b in deps
    }

    cg_path.unlink()
    cg_path.with_suffix('.pdf').unlink()


@pytest.mark.skipif(not graphviz_present(), reason='Graphviz is not installed')
def test_scheduler_typebound_ignore(tmp_path, testdir, config, frontend, proj_typebound_dependencies):
    """
    Test correct dependency chasing for typebound procedure calls with ignore working for
    typebound procedures correctly.

    projTypeBound: driver -> some_type%other_routine -> other_routine -> some_type%routine1 -> routine1
                   | | | | |                                          |                                |
                   | | | | |       +- routine <- some_type%routine2 <-+                                +---------+
                   | | | | |       |                                                                             |
                   | | | | +--> some_type%some_routine -> some_routine -> some_type%routine -> module_routine  <-+
                   | | +------> header_type%member_routine -> header_member_routine
                   | +--------> header_type%routine -> header_type%routine_real -> header_routine_real
                   |                           |
                   |                           +---> header_type%routine_integer -> routine_integer
                   +---------->other_type%member -> other_member -> header_member_routine
    """
    proj = testdir/'sources/projTypeBound'

    config['default']['disable'] += [
        'some_type%some_routine',
        'header_member_routine'
    ]
    config['routines'] = {
        'other_member': {
            'disable': config['default']['disable'] + ['member_routine']
        }
    }

    items_to_remove = (
        'typebound_item#some_type%some_routine',
        'typebound_item#some_routine',
        'typebound_item#some_type%routine',
        'typebound_header#header_member_routine',
    )

    proj_typebound_dependencies = {
        name: tuple(dep for dep in deps if dep not in items_to_remove)
        for name, deps in proj_typebound_dependencies.items()
        if name not in items_to_remove
    }

    scheduler = Scheduler(
        paths=proj, seed_routines=['driver'], config=config,
        full_parse=False, frontend=frontend, xmods=[tmp_path]
    )

    assert set(scheduler.items) == set(proj_typebound_dependencies)
    assert set(scheduler.dependencies) == {
        (a, b) for a, deps in proj_typebound_dependencies.items() for b in deps
    }

    cg_path = tmp_path/'callgraph_typebound'
    scheduler.callgraph(cg_path)

    # Testing of callgraph visualisation
    vgraph = VisGraphWrapper(cg_path)
    assert set(vgraph.nodes) == {n.upper() for n in proj_typebound_dependencies}
    assert set(vgraph.edges) == {
        (a.upper(), b.upper()) for a, deps in proj_typebound_dependencies.items() for b in deps
    }

    cg_path.unlink()
    cg_path.with_suffix('.pdf').unlink()


@pytest.mark.parametrize('frontend', available_frontends())
def test_scheduler_nested_type_enrichment(tmp_path, frontend, config):
    """
    Make sure that enrichment works correctly for nested types across
    multiple source files
    """
    fcode1 = """
module typebound_procedure_calls_mod
    implicit none

    type my_type
        integer :: val
    contains
        procedure :: reset
        procedure :: add => add_my_type
    end type my_type

    type other_type
        type(my_type) :: arr(3)
    contains
        procedure :: add => add_other_type
        procedure :: total_sum
    end type other_type

contains

    subroutine reset(this)
        class(my_type), intent(inout) :: this
        this%val = 0
    end subroutine reset

    subroutine add_my_type(this, val)
        class(my_type), intent(inout) :: this
        integer, intent(in) :: val
        this%val = this%val + val
    end subroutine add_my_type

    subroutine add_other_type(this, other)
        class(other_type) :: this
        type(other_type) :: other
        integer :: i
        do i=1,3
            call this%arr(i)%add(other%arr(i)%val)
        end do
    end subroutine add_other_type

    function total_sum(this) result(result)
        class(other_type), intent(in) :: this
        integer :: result
        integer :: i
        result = 0
        do i=1,3
            result = result + this%arr(i)%val
        end do
    end function total_sum

end module typebound_procedure_calls_mod
    """.strip()

    fcode2 = """
module other_typebound_procedure_calls_mod
    use typebound_procedure_calls_mod, only: other_type
    implicit none

    type third_type
        type(other_type) :: stuff(2)
    contains
        procedure :: init
        procedure :: print => print_content
    end type third_type

contains

    subroutine init(this)
        class(third_type), intent(inout) :: this
        integer :: i, j
        do i=1,2
            do j=1,3
                call this%stuff(i)%arr(j)%reset()
                call this%stuff(i)%arr(j)%add(i+j)
            end do
        end do
    end subroutine init

    subroutine print_content(this)
        class(third_type), intent(inout) :: this
        call this%stuff(1)%add(this%stuff(2))
        print *, this%stuff(1)%total_sum()
    end subroutine print_content
end module other_typebound_procedure_calls_mod
    """.strip()

    fcode3 = """
subroutine driver
    use other_typebound_procedure_calls_mod, only: third_type
    implicit none
    type(third_type) :: data
    integer :: mysum

    call data%init()
    call data%stuff(1)%arr(1)%add(1)
    mysum = data%stuff(1)%total_sum() + data%stuff(2)%total_sum()
    call data%print
end subroutine driver
    """.strip()

    (tmp_path/'typebound_procedure_calls_mod.F90').write_text(fcode1)
    (tmp_path/'other_typebound_procedure_calls_mod.F90').write_text(fcode2)
    (tmp_path/'driver.F90').write_text(fcode3)

    scheduler = Scheduler(
        paths=[tmp_path], config=config, seed_routines=['driver'],
        frontend=frontend, xmods=[tmp_path]
    )

    driver = scheduler['#driver'].source['driver']
    calls = FindNodes(ir.CallStatement).visit(driver.body)
    assert len(calls) == 3
    for call in calls:
        assert isinstance(call.name, ProcedureSymbol)
        assert isinstance(call.name.type.dtype, ProcedureType)
        assert call.name.parent
        assert isinstance(call.name.parent.type.dtype, DerivedType)
        assert isinstance(call.routine, Subroutine)
        assert isinstance(call.name.type.dtype.procedure, Subroutine)

    assert isinstance(calls[0].name.parent, Scalar)
    assert calls[0].name.parent.type.dtype.name == 'third_type'
    assert isinstance(calls[0].name.parent.type.dtype.typedef, ir.TypeDef)

    assert isinstance(calls[1].name.parent, Array)
    assert calls[1].name.parent.type.dtype.name == 'my_type'
    assert isinstance(calls[1].name.parent.type.dtype.typedef, ir.TypeDef)

    assert isinstance(calls[1].name.parent.parent, Array)
    assert isinstance(calls[1].name.parent.parent.type.dtype, DerivedType)
    assert calls[1].name.parent.parent.type.dtype.name == 'other_type'
    assert isinstance(calls[1].name.parent.parent.type.dtype.typedef, ir.TypeDef)

    assert isinstance(calls[1].name.parent.parent.parent, Scalar)
    assert isinstance(calls[1].name.parent.parent.parent.type.dtype, DerivedType)
    assert calls[1].name.parent.parent.parent.type.dtype.name == 'third_type'
    assert isinstance(calls[1].name.parent.parent.parent.type.dtype.typedef, ir.TypeDef)

    inline_calls = FindInlineCalls().visit(driver.body)
    assert len(inline_calls) == 2
    for call in inline_calls:
        assert isinstance(call.function, ProcedureSymbol)
        assert isinstance(call.function.type.dtype, ProcedureType)

        assert call.function.parent
        assert isinstance(call.function.parent, Array)
        assert isinstance(call.function.parent.type.dtype, DerivedType)
        assert call.function.parent.type.dtype.name == 'other_type'
        assert isinstance(call.function.parent.type.dtype.typedef, ir.TypeDef)

        assert call.function.parent.parent
        assert isinstance(call.function.parent.parent, Scalar)
        assert isinstance(call.function.parent.parent.type.dtype, DerivedType)
        assert call.function.parent.parent.type.dtype.name == 'third_type'
        assert isinstance(call.function.parent.parent.type.dtype.typedef, ir.TypeDef)


@pytest.mark.parametrize('frontend', available_frontends())
def test_scheduler_interface_inline_call(tmp_path, testdir, config, frontend):
    """
    Test that inline function calls declared via an explicit interface are added as dependencies.
    """

    my_config = config.copy()
    my_config['routines'] = {
        'driver': {
            'role': 'driver',
        }
    }

    scheduler = Scheduler(
        paths=testdir/'sources/projInlineCalls', config=my_config, frontend=frontend, xmods=[tmp_path]
    )

    expected_dependencies = {
        '#driver': (
            '#double_real', 'some_module', 'some_module#add_args', 'some_module#return_one',
            'some_module#some_type', 'some_module#some_type%do_something', 'vars_module',
        ),
        '#double_real': ('vars_module',),
        'some_module': (),
        'some_module#add_args': ('some_module#add_two_args', 'some_module#add_three_args'),
        'some_module#add_two_args': (),
        'some_module#add_three_args': (),
        'some_module#return_one': (),
        'some_module#some_type': (),
        'some_module#some_type%do_something': ('some_module#add_const',),
        'some_module#add_const': ('some_module#some_type',),
        'vars_module': (),
    }

    assert set(scheduler.items) == set(expected_dependencies)
    assert set(scheduler.dependencies) == {
        (a, b) for a, deps in expected_dependencies.items() for b in deps
    }

    assert isinstance(scheduler['some_module#add_args'], InterfaceItem)
    assert isinstance(scheduler['#double_real'], ProcedureItem)
    assert isinstance(scheduler['some_module#some_type'], TypeDefItem)
    assert isinstance(scheduler['some_module#add_two_args'], ProcedureItem)
    assert isinstance(scheduler['some_module#add_three_args'], ProcedureItem)

    cg_path = tmp_path/'callgraph'
    scheduler.callgraph(cg_path)

    # Testing of callgraph visualisation with imports
    vgraph = VisGraphWrapper(cg_path)
    assert set(vgraph.nodes) == {i.upper() for i in expected_dependencies}
    assert set(vgraph.edges) == {
        (a.upper(), b.upper()) for a, deps in expected_dependencies.items() for b in deps
    }


@pytest.mark.parametrize('frontend', available_frontends())
def test_scheduler_interface_dependencies(tmp_path, frontend, config):
    """
    Ensure that interfaces are treated as intermediate nodes and incur
    dependencies on the actual procedures
    """
    fcode_module = """
module test_scheduler_interface_dependencies_mod
    implicit none
    interface my_intf
        procedure proc1
        procedure proc2
    end interface my_intf
contains
    subroutine proc1(arg)
        integer, intent(inout) :: arg
        arg = arg + 1
    end subroutine proc1
    subroutine proc2(arg)
        real, intent(inout) :: arg
        arg = arg + 1.0
    end subroutine proc2
end module test_scheduler_interface_dependencies_mod
    """
    fcode_driver = """
subroutine test_scheduler_interface_dependencies_driver
    use test_scheduler_interface_dependencies_mod, only: my_intf
    implicit none
    integer i
    real a
    i = 0
    a = 0.0
    call my_intf(i)
    call my_intf(a)
end subroutine test_scheduler_interface_dependencies_driver
    """

    config['routines']['test_scheduler_interface_dependencies_driver'] = {
        'role': 'driver'
    }

    (tmp_path/'test_scheduler_interface_dependencies_mod.F90').write_text(fcode_module)
    (tmp_path/'test_scheduler_interface_dependencies_driver.F90').write_text(fcode_driver)

    scheduler = Scheduler(
        paths=[tmp_path], config=config, seed_routines=['test_scheduler_interface_dependencies_driver'],
        frontend=frontend, xmods=[tmp_path]
    )

    expected_dependencies = {
        '#test_scheduler_interface_dependencies_driver': {
            'test_scheduler_interface_dependencies_mod#my_intf'
        },
        'test_scheduler_interface_dependencies_mod#my_intf': {
            'test_scheduler_interface_dependencies_mod#proc1', 'test_scheduler_interface_dependencies_mod#proc2'
        },
        'test_scheduler_interface_dependencies_mod#proc1': set(),
        'test_scheduler_interface_dependencies_mod#proc2': set()
    }

    assert set(scheduler.items) == set(expected_dependencies)
    assert set(scheduler.dependencies) == {
        (a, b) for a, deps in expected_dependencies.items() for b in deps
    }

    assert isinstance(scheduler['test_scheduler_interface_dependencies_mod#my_intf'], InterfaceItem)
    assert isinstance(scheduler['test_scheduler_interface_dependencies_mod#proc1'], ProcedureItem)
    assert isinstance(scheduler['test_scheduler_interface_dependencies_mod#proc2'], ProcedureItem)


@pytest.mark.parametrize('full_parse', [True, False])
def test_scheduler_typebound_inline_call(tmp_path, config, full_parse):
    fcode_mod = """
module some_mod
    implicit none
    type some_type
        integer :: a
    contains
        procedure :: some_routine
        procedure :: some_function
    end type some_type
contains
    subroutine some_routine(t)
        class(some_type), intent(inout) :: t
        t%a = 5
    end subroutine some_routine

    integer function some_function(t)
        class(some_type), intent(in) :: t
        some_function = t%a
    end function some_function
end module some_mod
    """.strip()

    fcode_caller = """
subroutine caller(b)
    use some_mod, only: some_type
    implicit none
    integer, intent(inout) :: b
    type(some_type) :: t
    t%a = b
    call t%some_routine()
    b = t%some_function()
end subroutine caller
    """.strip()

    (tmp_path/'some_mod.F90').write_text(fcode_mod)
    (tmp_path/'caller.F90').write_text(fcode_caller)

    def verify_graph(scheduler, expected_dependencies):
        assert set(scheduler.items) == set(expected_dependencies)
        assert set(scheduler.dependencies) == {
            (a, b) for a, deps in expected_dependencies.items() for b in deps
        }

        assert all(item.source._incomplete is not full_parse for item in scheduler.items)

        cg_path = tmp_path/'callgraph'
        scheduler.callgraph(cg_path)

        # Testing of callgraph visualisation
        vgraph = VisGraphWrapper(cg_path)
        assert set(vgraph.nodes) == {n.upper() for n in expected_dependencies}
        assert set(vgraph.edges) == {
            (a.upper(), b.upper()) for a, deps in expected_dependencies.items()
            for b in deps
        }

    scheduler = Scheduler(
        paths=[tmp_path], config=config, seed_routines=['caller'], full_parse=full_parse, xmods=[tmp_path]
    )

    expected_dependencies = {
        '#caller': (
            'some_mod#some_type',
            'some_mod#some_type%some_routine',
        ),
        'some_mod#some_type': (),
        'some_mod#some_type%some_routine': ('some_mod#some_routine',),
        'some_mod#some_routine': ('some_mod#some_type',),
    }

    if scheduler.full_parse:
        # Inline Calls can only be fully resolved in a full parse
        expected_dependencies['#caller'] += ('some_mod#some_type%some_function',)
        expected_dependencies['some_mod#some_type%some_function'] = ('some_mod#some_function',)
        expected_dependencies['some_mod#some_function'] = ('some_mod#some_type',)

    verify_graph(scheduler, expected_dependencies)


@pytest.mark.parametrize('full_parse', [False, True])
def test_scheduler_cycle(tmp_path, config, full_parse):
    fcode_mod = """
module some_mod
    implicit none
    type some_type
        integer :: a
    contains
        procedure :: proc => some_proc
        procedure :: other => some_other
    end type some_type
contains
    recursive subroutine some_proc(this, val, recurse, fallback)
        class(some_type), intent(inout) :: this
        integer, intent(in) :: val
        logical, intent(in), optional :: recurse

        if (present(recurse)) then
            if (present(fallback)) then
                call this%other(val)
            else
                call some_proc(this, val, .true., .true.)
            end if
        else
            call this%proc(val, .true.)
        end if
    end subroutine some_proc

    subroutine some_other(this, val)
        class(some_type), intent(inout) :: this
        integer, intent(in) :: val
        this%a = val
    end subroutine some_other
end module some_mod
    """.strip()

    fcode_caller = """
subroutine caller
    use some_mod, only: some_type
    implicit none
    type(some_type) :: t

    call t%proc(1)
end subroutine caller
    """.strip()

    (tmp_path/'some_mod.F90').write_text(fcode_mod)
    (tmp_path/'caller.F90').write_text(fcode_caller)

    scheduler = Scheduler(
        paths=[tmp_path], config=config, seed_routines=['caller'], full_parse=full_parse, xmods=[tmp_path]
    )

    # Make sure we the outgoing edges from the recursive routine to the procedure binding
    # and itself are removed but the other edge still exists
    assert (scheduler['#caller'], scheduler['some_mod#some_type%proc']) in scheduler.dependencies
    assert (scheduler['some_mod#some_type%proc'], scheduler['some_mod#some_proc']) in scheduler.dependencies
    assert (scheduler['some_mod#some_proc'], scheduler['some_mod#some_type%proc']) not in scheduler.dependencies
    assert (scheduler['some_mod#some_proc'], scheduler['some_mod#some_proc']) not in scheduler.dependencies
    assert (scheduler['some_mod#some_proc'], scheduler['some_mod#some_type%other']) in scheduler.dependencies
    assert (scheduler['some_mod#some_type%other'], scheduler['some_mod#some_other']) in scheduler.dependencies


def test_scheduler_unqualified_imports(config):
    """
    Test that only qualified imports are added as children.
    """

    kernel = """
    subroutine kernel()
       use some_mod
       use other_mod, only: other_routine

       call other_routine
    end subroutine kernel
    """

    source = Sourcefile.from_source(kernel, frontend=REGEX)
    item = ProcedureItem(name='#kernel', source=source, config=config['default'])

    assert len(item.dependencies) == 3
    children = set()
    for dep in item.dependencies:
        if isinstance(dep, ir.Import):
            if dep.symbols:
                children |= {f'{dep.module}#{str(s)}'.lower() for s in dep.symbols}
            else:
                children.add(dep.module.lower())
        elif isinstance(dep, ir.CallStatement):
            children.add(str(dep.name).lower())
        else:
            assert False, 'Unexpected dependency type'
    assert children == {'some_mod', 'other_mod#other_routine', 'other_routine'}


@pytest.mark.parametrize('frontend', available_frontends(skip={OMNI: "OMNI fails on missing module"}))
@pytest.mark.parametrize('enable_imports', [False, True])
@pytest.mark.parametrize('import_level', ['module', 'subroutine'])
def test_scheduler_indirect_import(frontend, tmp_path, enable_imports, import_level):
    fcode_mod_a = """
module a_mod
    implicit none
    public
    integer :: global_a = 1
end module a_mod
"""

    fcode_mod_b = """
module b_mod
    use a_mod
    implicit none
    public
    type type_b
        integer :: val
    end type type_b
end module b_mod
"""

    module_import_stmt = ""
    routine_import_stmt = ""
    if import_level == 'module':
        module_import_stmt = "use b_mod, only: type_b, global_a"
    elif import_level == 'subroutine':
        routine_import_stmt = "use b_mod, only: type_b, global_a"

    fcode_mod_c = f"""
module c_mod
    {module_import_stmt}
    implicit none
contains
    subroutine c(b)
        {routine_import_stmt}
        implicit none
        type(type_b), intent(inout) :: b
        b%val = global_a
    end subroutine c
end module c_mod
"""

    # Set-up paths and write sources
    src_path = tmp_path/'src'
    src_path.mkdir()
    out_path = tmp_path/'build'
    out_path.mkdir()

    (src_path/'a.F90').write_text(fcode_mod_a)
    (src_path/'b.F90').write_text(fcode_mod_b)
    (src_path/'c.F90').write_text(fcode_mod_c)

    config = SchedulerConfig.from_dict({
        'default': {
            'role': 'kernel',
            'expand': True,
            'strict': True,
            'enable_imports': enable_imports
        },
        'routines': {'c': {'role': 'driver'}}
    })
    # Create the Scheduler
    scheduler = Scheduler(
        paths=[src_path], config=config, frontend=frontend,
        output_dir=out_path, xmods=[out_path]
    )

    # Check for all items in the dependency graph
    expected_items = {'a_mod', 'b_mod', 'b_mod#type_b', 'c_mod#c'}
    assert expected_items == {item.name for item in scheduler.items}

    type_b = scheduler['b_mod#type_b'].ir
    c_mod_c = scheduler['c_mod#c'].ir
    var_map = CaseInsensitiveDict(
        (v.name, v) for v in FindVariables().visit(c_mod_c.body)
    )
    global_a = var_map['global_a']
    b_dtype = var_map['b'].type.dtype

    # Verify the type information for the imported symbols:
    # They will have enriched information if the imports are enabled
    # and deferred type otherwise
    if enable_imports:
        assert global_a.type.dtype is BasicType.INTEGER
        assert global_a.type.initial == '1'
        assert b_dtype.typedef is type_b
    else:
        assert global_a.type.dtype is BasicType.DEFERRED
        assert global_a.type.initial is None
        assert b_dtype.typedef is BasicType.DEFERRED


@pytest.mark.parametrize('frontend', available_frontends(skip={OMNI: "OMNI fails on missing module"}))
@pytest.mark.parametrize('external_kernel', [None, 'module', 'intfb'])
def test_scheduler_ignore_external_item(frontend, tmp_path, external_kernel):
    fcode_driver = f"""
module driver_mod
  contains
  subroutine driver(nlon, klev, nb, ydml_phy_mf)
    use parkind1, only: jpim, jprb
    use kernel1_mod, only: kernel1
    {'use kernel2_mod, only: kernel2' if external_kernel == 'module' else ''}
    implicit none
    type(model_physics_mf_type), intent(in) :: ydml_phy_mf
    integer(kind=jpim), intent(in) :: nlon
    integer(kind=jpim), intent(in) :: klev
    integer(kind=jpim), intent(in) :: nb
    integer(kind=jpim) :: jstart
    integer(kind=jpim) :: jend
    integer(kind=jpim) :: b
{'#include "kernel2.intfb.h"' if external_kernel == 'intfb' else ''}
    jstart = 1
    jend = nlon
    do b = 1, nb
        call kernel1()
        {'call kernel2()' if external_kernel else ''}
    enddo
  end subroutine driver
end module driver_mod
    """.strip()
    fcode_kernel1 = """
module kernel1_mod
  contains
  subroutine kernel1()
    use parkind1, only: jpim, jprb
  end subroutine kernel1
end module kernel1_mod
    """.strip()

    (tmp_path/'driver.F90').write_text(fcode_driver)
    (tmp_path/'kernel1_mod.F90').write_text(fcode_kernel1)

    config = {
        'default': {
            'mode': 'idem',
            'role': 'kernel',
            'expand': True,
            'strict': True,
            'ignore': ['parkind1'],
        },
        'routines': {
            'driver': {'role': 'driver'}
        }
    }

    if external_kernel:
        config['default']['generated'] = ['kernel2*']

    class Trafo(Transformation):

        item_filter = (ProcedureItem, ModuleItem)

        def transform_module(self, module, **kwargs):
            pass

        def transform_subroutine(self, routine, **kwargs):
            pass

    scheduler = Scheduler(
        paths=[tmp_path], config=SchedulerConfig.from_dict(config),
        definitions=(), xmods=[tmp_path], frontend=frontend
    )

    expected_items = {'driver_mod#driver', 'parkind1', 'kernel1_mod#kernel1'}
    if external_kernel == 'module':
        expected_items.add('kernel2_mod#kernel2')
        expected_items.add('kernel2_mod')
    elif external_kernel == 'intfb':
        expected_items.add('#kernel2')

    assert expected_items == {item.name for item in scheduler.items}
    for item in scheduler.items:
        if item.name == 'parkind1':
            assert item.is_ignored
            assert isinstance(item, ExternalItem)
        if item.name.endswith('#kernel2'):
            assert not item.is_ignored
            assert isinstance(item, ExternalItem)
            assert item.is_generated

    if external_kernel:
        scheduler.process(transformation=Trafo(), proc_strategy=ProcessingStrategy.PLAN)
        # this shouldn't fail because we marked the item as build-time generated
        with pytest.raises(RuntimeError):
            scheduler.process(transformation=Trafo())
    else:
        # check whether this works without any error
        scheduler.process(transformation=Trafo())


@pytest.mark.parametrize('frontend', available_frontends())
@pytest.mark.parametrize('qualified_import', ['driver', 'module', 'both', 'none'])
def test_scheduler_transient_typedef_imports(frontend, qualified_import, tmp_path):
    """ Test that use of transiently imported typedefs succeeds. """

    fcode_mod_a = f"""
module mod_a
    use mod_b{', only: my_type' if qualified_import in ('module', 'both') else ''}
    implicit none
end module mod_a
"""

    fcode_mod_b = """
module mod_b
    implicit none
    type my_type
        real(kind=4) :: a, b, x

        contains
        procedure :: add_a_b => my_type_add_a_b
    end type my_type

contains
    subroutine my_type_add_a_b(obj)
        type(my_type), intent(inout) :: obj

        obj%x = obj%a + obj%b
    end subroutine my_type_add_a_b
end module mod_b
"""

    fcode_driver = f"""
subroutine test_scheduler()
    use mod_a {', only: my_type' if qualified_import in ('driver', 'both') else ''}
    implicit none

    type(my_type) :: d

    d%a = 42.0
    d%a = 66.6
    call d%add_a_b()
end subroutine test_scheduler
"""
    src_path = tmp_path/'src'
    src_path.mkdir()
    (src_path/'mod_a.F90').write_text(fcode_mod_a)
    (src_path/'mod_b.F90').write_text(fcode_mod_b)
    (src_path/'driver.F90').write_text(fcode_driver)

    config = SchedulerConfig.from_dict({
        'default': {
            'role': 'kernel', 'expand': True, 'strict': True, 'enable_imports': True
        },
        'routines': {'test_scheduler': {'role': 'driver'}}
    })
    # Create the Scheduler
    if qualified_import not in ('driver', 'both'):
        with pytest.raises(RuntimeError):
            _ = Scheduler(
                paths=[src_path], config=config, seed_routines='test_scheduler',
                frontend=frontend, xmods=[tmp_path], full_parse=False
            )
    else:
        scheduler = Scheduler(
            paths=[src_path], config=config, seed_routines='test_scheduler',
            frontend=frontend, xmods=[tmp_path], full_parse=False
        )

        if qualified_import == 'both':
            assert scheduler.items == (
                '#test_scheduler', 'mod_a', 'mod_a#my_type%add_a_b', 'mod_b#my_type'
            )
            assert scheduler.dependencies == (
                ('#test_scheduler', 'mod_a'),
                ('#test_scheduler', 'mod_a#my_type%add_a_b'),
                ('mod_a', 'mod_b#my_type')
            )
        else:
            assert scheduler.items == (
                '#test_scheduler', 'mod_a', 'mod_a#my_type%add_a_b', 'mod_b'
            )
            assert scheduler.dependencies == (
                ('#test_scheduler', 'mod_a'),
                ('#test_scheduler', 'mod_a#my_type%add_a_b'),
                ('mod_a', 'mod_b')
            )

        if frontend == OMNI:
            # OMNI fails to read due to missing mod_b xmods
            with pytest.raises(ParseError):
                scheduler._parse_items()
        else:
            scheduler._parse_items()

        call = FindNodes(ir.CallStatement).visit(scheduler['#test_scheduler'].ir.ir)[0]
        assert call.name == 'd%add_a_b'
        assert isinstance(call.name.parent.type.dtype, DerivedType)
        assert call.name.parent.type.dtype.name == 'my_type'

        # Enrichment does work correctly in this situation, providing the link to the typedef
        if qualified_import in ('driver', 'both') and frontend == FP:
            assert call.name.parent.type.dtype.typedef is (
                getattr(scheduler['mod_b#my_type'], 'ir', None) or scheduler['mod_b'].ir['my_type']
            )
        else:
            # The interprocedural annotations do _not_ currently enrich the type in this situation
            assert call.name.parent.type.dtype.typedef == BasicType.DEFERRED


@pytest.mark.parametrize('frontend', available_frontends())
@pytest.mark.parametrize('qualified_import', ['driver', 'module', 'both', 'none'])
def test_scheduler_transient_procedure_imports(frontend, qualified_import, tmp_path):
    """ Test that use of transiently imported procedures succeeds. """

    fcode_mod_a = f"""
module mod_a
    use mod_b{', only: my_proc' if qualified_import in ('module', 'both') else ''}
    implicit none
end module mod_a
"""

    fcode_mod_b = """
module mod_b
    implicit none
contains
    subroutine my_proc
        print *,'hello world'
    end subroutine
end module mod_b
"""

    fcode_driver = f"""
subroutine test_scheduler()
    use mod_a {', only: my_proc' if qualified_import in ('driver', 'both') else ''}
    implicit none

    call my_proc
end subroutine test_scheduler
"""
    src_path = tmp_path/'src'
    src_path.mkdir()
    (src_path/'mod_a.F90').write_text(fcode_mod_a)
    (src_path/'mod_b.F90').write_text(fcode_mod_b)
    (src_path/'driver.F90').write_text(fcode_driver)

    config = SchedulerConfig.from_dict({
        'default': {
            'role': 'kernel', 'expand': True, 'strict': True, 'enable_imports': True
        },
        'routines': {'test_scheduler': {'role': 'driver'}}
    })

    # Create the Scheduler
    if qualified_import in ('module', 'none'):
        with pytest.raises(RuntimeError):
            _ = Scheduler(
                paths=[src_path], config=config, seed_routines='test_scheduler',
                frontend=frontend, xmods=[tmp_path], full_parse=False
            )
    else:
        scheduler = Scheduler(
            paths=[src_path], config=config, seed_routines='test_scheduler',
            frontend=frontend, xmods=[tmp_path], full_parse=False
        )

        if qualified_import == 'both':
            assert scheduler.items == ('#test_scheduler', 'mod_a', 'mod_a#my_proc')
            assert scheduler.dependencies == (
                ('#test_scheduler', 'mod_a'), ('#test_scheduler', 'mod_a#my_proc')
            )
        else:
            assert scheduler.items == ('#test_scheduler', 'mod_a', 'mod_a#my_proc', 'mod_b')
            assert scheduler.dependencies == (
                ('#test_scheduler', 'mod_a'), ('#test_scheduler', 'mod_a#my_proc'),
                ('mod_a', 'mod_b')
            )

        if frontend == OMNI and qualified_import == 'both':
            # OMNI fails to read due to missing mod_b xmods
            with pytest.raises(Exception):
                scheduler._parse_items()
        else:
            scheduler._parse_items()

        call = FindNodes(ir.CallStatement).visit(scheduler['#test_scheduler'].ir.ir)[0]
        # The interprocedural annotations do _not_ currently enrich the call in this situation
        assert call.name == 'my_proc'
        assert call.name.type.imported
        assert not call.name.type.procedure


@pytest.mark.parametrize('frontend', available_frontends())
@pytest.mark.parametrize('qualified_import', ['driver', 'module', 'both', 'none'])
def test_scheduler_transient_variable_imports(frontend, qualified_import, tmp_path):
    """ Test that use of transiently imported variables succeeds. """

    fcode_mod_a = f"""
module mod_a
    use mod_b{', only: my_var' if qualified_import in ('module', 'both') else ''}
    implicit none
end module mod_a
"""

    fcode_mod_b = """
module mod_b
    implicit none
    integer :: my_var
end module mod_b
"""

    fcode_driver = f"""
subroutine test_scheduler()
    use mod_a {', only: my_var' if qualified_import in ('driver', 'both') else ''}
    implicit none
    my_var = 1
end subroutine test_scheduler
"""
    src_path = tmp_path/'src'
    src_path.mkdir()
    (src_path/'mod_a.F90').write_text(fcode_mod_a)
    (src_path/'mod_b.F90').write_text(fcode_mod_b)
    (src_path/'driver.F90').write_text(fcode_driver)

    config = SchedulerConfig.from_dict({
        'default': {
            'role': 'kernel', 'expand': True, 'strict': True, 'enable_imports': True
        },
        'routines': {'test_scheduler': {'role': 'driver'}}
    })

    scheduler = Scheduler(
        paths=[src_path], config=config, seed_routines='test_scheduler',
        frontend=frontend, xmods=[tmp_path], full_parse=False
    )

    assert scheduler.items == ('#test_scheduler', 'mod_a', 'mod_b')
    assert scheduler.dependencies == (
        ('#test_scheduler', 'mod_a'), ('mod_a', 'mod_b')
    )

    scheduler._parse_items()

    my_var = scheduler['#test_scheduler'].ir.body.body[0].lhs
    # NB: Global variable imports as a dependency are established directly on the import
    #     statements, thus the transient dependency is captured
    assert isinstance(my_var, Scalar)
    assert my_var.type.dtype == BasicType.INTEGER
    assert my_var.type.imported
    assert my_var.type.module is scheduler['mod_a'].ir

    if qualified_import in ('module', 'both'):
        mod_a_var = scheduler['mod_a'].ir.imported_symbol_map['my_var']
        # NB: We are able to correctly propagate the type information through
        #     multiple import layers
        assert isinstance(mod_a_var, Scalar)
        assert mod_a_var.type.dtype == BasicType.INTEGER
        assert mod_a_var.type.imported
        # NB: This links to the module it imports from, not where it is defined
        assert mod_a_var.type.module is scheduler['mod_b'].ir


def test_scheduler_module_interface_import(frontend, tmp_path):
    """ Test module-level imports of interface routines. """

    fcode_mod_a = """
module mod_a
    use mod_b, only: inner_type, intfb_routine
    implicit none
    type outer_type
        type(inner_type) :: da
    end type outer_type
end module mod_a
"""

    fcode_mod_b = """
module mod_b
    implicit none
    type inner_type
        integer :: da, db
    end type inner_type

    interface intfb_routine
    module procedure routine_a, routine_b
    end interface intfb_routine

    contains
    subroutine routine_a
    end subroutine routine_a

    subroutine routine_b
    end subroutine routine_b
end module mod_b
"""

    fcode_driver = """
subroutine test_scheduler()
    use mod_a, only: outer_type
    implicit none
    type(outer_type) :: da

    da%da%da = 42.0
end subroutine test_scheduler
"""
    src_path = tmp_path/'src'
    src_path.mkdir()
    (src_path/'mod_a.F90').write_text(fcode_mod_a)
    (src_path/'mod_b.F90').write_text(fcode_mod_b)
    (src_path/'driver.F90').write_text(fcode_driver)

    config = SchedulerConfig.from_dict({
        'default': {
            'role': 'kernel', 'expand': True, 'strict': True, 'enable_imports': True
        },
        'routines': {'test_scheduler': {'role': 'driver'}}
    })

    scheduler = Scheduler(
        paths=[src_path], config=config, seed_routines='test_scheduler',
        frontend=frontend, xmods=[tmp_path], full_parse=False
    )

    assert scheduler.items == (
        '#test_scheduler', 'mod_a#outer_type', 'mod_b#inner_type'
    )
    assert scheduler.dependencies == (
        ('#test_scheduler', 'mod_a#outer_type'),
        ('mod_a#outer_type', 'mod_b#inner_type')
    )
